import asyncio
import importlib
import json
import logging
import os
import sys
from typing import TYPE_CHECKING

import aiohttp
import psutil

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


class Reporter:
    def __init__(self, config, trainer, cm: "CommunicationsManager"):
        logging.info("Starting reporter module")
        self.config = config
        self.trainer = trainer
        self.cm = cm
        self.frequency = self.config.participant["reporter_args"]["report_frequency"]
        self.grace_time = self.config.participant["reporter_args"]["grace_time_reporter"]
        self.data_queue = asyncio.Queue()
        self.url = f"http://{self.config.participant['scenario_args']['controller']}/nebula/dashboard/{self.config.participant['scenario_args']['name']}/node/update"
        self.counter = 0

        self.first_net_metrics = True
        self.prev_bytes_sent = 0
        self.prev_bytes_recv = 0
        self.prev_packets_sent = 0
        self.prev_packets_recv = 0

        self.acc_bytes_sent = 0
        self.acc_bytes_recv = 0
        self.acc_packets_sent = 0
        self.acc_packets_recv = 0

    async def enqueue_data(self, name, value):
        await self.data_queue.put((name, value))

    async def start(self):
        await asyncio.sleep(self.grace_time)
        asyncio.create_task(self.run_reporter())

    async def run_reporter(self):
        while True:
            # NOTE: currently disabled
            # if self.config.participant["scenario_args"]["controller"] != "nebula-test":
            #     await self.__report_status_to_controller()
            # await self.__report_data_queue()
            await self.__report_resources()
            self.counter += 1
            if self.counter % 50 == 0:
                logging.info("Reloading config file...")
                self.cm.engine.config.reload_config_file()
            await asyncio.sleep(self.frequency)

    async def report_scenario_finished(self):
        url = f"http://{self.config.participant['scenario_args']['controller']}/nebula/dashboard/{self.config.participant['scenario_args']['name']}/node/done"
        data = json.dumps({"idx": self.config.participant["device_args"]["idx"]})
        headers = {
            "Content-Type": "application/json",
            "User-Agent": f"NEBULA Participant {self.config.participant['device_args']['idx']}",
        }
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(url, data=data, headers=headers) as response:
                    if response.status != 200:
                        logging.error(
                            f"Error received from controller: {response.status} (probably there is overhead in the controller, trying again in the next round)"
                        )
                        text = await response.text()
                        logging.debug(text)
                    else:
                        logging.info(
                            f"Participant {self.config.participant['device_args']['idx']} reported scenario finished"
                        )
                        return True
        except aiohttp.ClientError as e:
            logging.exception(f"Error connecting to the controller at {url}: {e}")
        return False

    async def __report_data_queue(self):
        while not self.data_queue.empty():
            name, value = await self.data_queue.get()
            await self.trainer.logger.log_data({name: value})  # Assuming log_data can be made async
            self.data_queue.task_done()

    async def __report_status_to_controller(self):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    self.url,
                    data=json.dumps(self.config.participant),
                    headers={
                        "Content-Type": "application/json",
                        "User-Agent": f"NEBULA Participant {self.config.participant['device_args']['idx']}",
                    },
                ) as response:
                    if response.status != 200:
                        logging.error(
                            f"Error received from controller: {response.status} (probably there is overhead in the controller, trying again in the next round)"
                        )
                        text = await response.text()
                        logging.debug(text)
        except aiohttp.ClientError as e:
            logging.exception(f"Error connecting to the controller at {self.url}: {e}")
        except Exception as e:
            logging.exception(f"Error sending status to controller, will try again in a few seconds: {e}")
            await asyncio.sleep(5)

    async def __report_resources(self):
        cpu_percent = psutil.cpu_percent()
        cpu_temp = 0
        try:
            if sys.platform == "linux":
                sensors = await asyncio.to_thread(psutil.sensors_temperatures)
                cpu_temp = sensors.get("coretemp")[0].current if sensors.get("coretemp") else 0
        except Exception:
            pass

        pid = os.getpid()
        cpu_percent_process = await asyncio.to_thread(psutil.Process(pid).cpu_percent, interval=1)

        process = psutil.Process(pid)
        memory_process = await asyncio.to_thread(lambda: process.memory_info().rss / (1024**2))
        memory_percent_process = process.memory_percent()
        memory_info = await asyncio.to_thread(psutil.virtual_memory)
        memory_percent = memory_info.percent
        memory_used = memory_info.used / (1024**2)

        disk_percent = psutil.disk_usage("/").percent

        net_io_counters = await asyncio.to_thread(psutil.net_io_counters)
        bytes_sent = net_io_counters.bytes_sent
        bytes_recv = net_io_counters.bytes_recv
        packets_sent = net_io_counters.packets_sent
        packets_recv = net_io_counters.packets_recv

        if self.first_net_metrics:
            bytes_sent_diff = 0
            bytes_recv_diff = 0
            packets_sent_diff = 0
            packets_recv_diff = 0
            self.first_net_metrics = False
        else:
            bytes_sent_diff = bytes_sent - self.prev_bytes_sent
            bytes_recv_diff = bytes_recv - self.prev_bytes_recv
            packets_sent_diff = packets_sent - self.prev_packets_sent
            packets_recv_diff = packets_recv - self.prev_packets_recv

        self.prev_bytes_sent = bytes_sent
        self.prev_bytes_recv = bytes_recv
        self.prev_packets_sent = packets_sent
        self.prev_packets_recv = packets_recv

        self.acc_bytes_sent += bytes_sent_diff
        self.acc_bytes_recv += bytes_recv_diff
        self.acc_packets_sent += packets_sent_diff
        self.acc_packets_recv += packets_recv_diff

        current_connections = await self.cm.get_addrs_current_connections(only_direct=True)

        resources = {
            "CPU/CPU global (%)": cpu_percent,
            "CPU/CPU process (%)": cpu_percent_process,
            "CPU/CPU temperature (°)": cpu_temp,
            "RAM/RAM global (%)": memory_percent,
            "RAM/RAM global (MB)": memory_used,
            "RAM/RAM process (%)": memory_percent_process,
            "RAM/RAM process (MB)": memory_process,
            "Disk/Disk (%)": disk_percent,
            "Network/Network (bytes sent)": round(self.acc_bytes_sent / (1024**2), 3),
            "Network/Network (bytes received)": round(self.acc_bytes_recv / (1024**2), 3),
            "Network/Network (packets sent)": self.acc_packets_sent,
            "Network/Network (packets received)": self.acc_packets_recv,
            "Network/Connections": len(current_connections),
        }
        self.trainer.logger.log_data(resources)

        if importlib.util.find_spec("pynvml") is not None:
            try:
                import pynvml

                await asyncio.to_thread(pynvml.nvmlInit)
                devices = await asyncio.to_thread(pynvml.nvmlDeviceGetCount)
                for i in range(devices):
                    handle = await asyncio.to_thread(pynvml.nvmlDeviceGetHandleByIndex, i)
                    gpu_percent = (await asyncio.to_thread(pynvml.nvmlDeviceGetUtilizationRates, handle)).gpu
                    gpu_temp = await asyncio.to_thread(
                        pynvml.nvmlDeviceGetTemperature,
                        handle,
                        pynvml.NVML_TEMPERATURE_GPU,
                    )
                    gpu_mem = await asyncio.to_thread(pynvml.nvmlDeviceGetMemoryInfo, handle)
                    gpu_mem_percent = round(gpu_mem.used / gpu_mem.total * 100, 3)
                    gpu_power = await asyncio.to_thread(pynvml.nvmlDeviceGetPowerUsage, handle) / 1000.0
                    gpu_clocks = await asyncio.to_thread(pynvml.nvmlDeviceGetClockInfo, handle, pynvml.NVML_CLOCK_SM)
                    gpu_memory_clocks = await asyncio.to_thread(
                        pynvml.nvmlDeviceGetClockInfo, handle, pynvml.NVML_CLOCK_MEM
                    )
                    gpu_fan_speed = await asyncio.to_thread(pynvml.nvmlDeviceGetFanSpeed, handle)
                    gpu_info = {
                        f"GPU/GPU{i} (%)": gpu_percent,
                        f"GPU/GPU{i} temperature (°)": gpu_temp,
                        f"GPU/GPU{i} memory (%)": gpu_mem_percent,
                        f"GPU/GPU{i} power": gpu_power,
                        f"GPU/GPU{i} clocks": gpu_clocks,
                        f"GPU/GPU{i} memory clocks": gpu_memory_clocks,
                        f"GPU/GPU{i} fan speed": gpu_fan_speed,
                    }
                    self.trainer.logger.log_data(gpu_info)
            except Exception:
                pass



class Role:
    """
    This class defines the participant roles of the platform.
    """

    TRAINER = "trainer"
    AGGREGATOR = "aggregator"
    PROXY = "proxy"
    IDLE = "idle"
    SERVER = "server"



import math

import torch

from nebula.core.models.nebulamodel import NebulaModel


class Sentiment140ModelCNN(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}
        self.example_input_array = torch.zeros(1, 1, 28, 28)
        self.learning_rate = learning_rate

        self.criterion = torch.nn.CrossEntropyLoss()

        self.filter_sizes = [2, 3, 4]
        self.n_filters = math.ceil(300 * len(self.filter_sizes) / 3)
        self.convs = torch.nn.ModuleList([
            torch.nn.Conv2d(in_channels=1, out_channels=self.n_filters, kernel_size=(fs, 300))
            for fs in self.filter_sizes
        ])
        self.fc = torch.nn.Linear(len(self.filter_sizes) * self.n_filters, self.num_classes)
        self.dropout = torch.nn.Dropout(0.5)

        self.epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}

    def forward(self, x):
        x = x.unsqueeze(1)
        conved = [
            torch.nn.functional.relu(conv(x)).squeeze(3) for conv in self.convs
        ]  # [(batch_size, n_filters, sent_len), ...] * len(filter_sizes)
        pooled = [
            torch.nn.functional.max_pool1d(conv, conv.shape[2]).squeeze(2) for conv in conved
        ]  # [(batch_size, n_filters), ...] * len(filter_sizes)
        cat = self.dropout(torch.cat(pooled, dim=1))
        out = self.fc(cat)
        return out

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.learning_rate,
            betas=(self.config["beta1"], self.config["beta2"]),
            amsgrad=self.config["amsgrad"],
        )
        return optimizer



import asyncio
import logging
from abc import ABC, abstractmethod
from collections import deque
from typing import TYPE_CHECKING, Any

from nebula.addons.functions import print_msg_box

if TYPE_CHECKING:
    from nebula.config.config import Config
    from nebula.core.aggregation.aggregator import Aggregator
    from nebula.core.engine import Engine
    from nebula.core.network.communications import CommunicationsManager
    from nebula.core.training.lightning import Lightning


class PropagationStrategy(ABC):
    @abstractmethod
    def is_node_eligible(self, node: str) -> bool:
        pass

    @abstractmethod
    def prepare_model_payload(self, node: str) -> tuple[Any, float] | None:
        pass


class InitialModelPropagation(PropagationStrategy):
    def __init__(self, aggregator: "Aggregator", trainer: "Lightning", engine: "Engine"):
        self.aggregator = aggregator
        self.trainer = trainer
        self.engine = engine

    def get_round(self):
        return self.engine.get_round()

    def is_node_eligible(self, node: str) -> bool:
        return node not in self.engine.cm.get_ready_connections()

    def prepare_model_payload(self, node: str) -> tuple[Any, float] | None:
        return (
            self.trainer.get_model_parameters(initialize=True),
            self.trainer.DEFAULT_MODEL_WEIGHT,
        )


class StableModelPropagation(PropagationStrategy):
    def __init__(self, aggregator: "Aggregator", trainer: "Lightning", engine: "Engine"):
        self.aggregator = aggregator
        self.trainer = trainer
        self.engine = engine
        self.addr = self.engine.get_addr()

    def get_round(self):
        return self.engine.get_round()

    def is_node_eligible(self, node: str) -> bool:
        return (node not in self.aggregator.get_nodes_pending_models_to_aggregate()) or (
            self.engine.cm.connections[node].get_federated_round() < self.get_round()
        )

    def prepare_model_payload(self, node: str) -> tuple[Any, float] | None:
        return self.trainer.get_model_parameters(), self.trainer.get_model_weight()


class Propagator:
    def __init__(self, cm: "CommunicationsManager"):
        self.engine: Engine = cm.engine
        self.config: Config = cm.get_config()
        self.addr = cm.get_addr()
        self.cm: CommunicationsManager = cm
        self.aggregator: Aggregator = self.engine.aggregator
        self.trainer: Lightning = self.engine._trainer

        self.status_history = deque(maxlen=self.config.participant["propagator_args"]["history_size"])

        self.interval = self.config.participant["propagator_args"]["propagate_interval"]
        self.model_interval = self.config.participant["propagator_args"]["propagate_model_interval"]
        self.early_stop = self.config.participant["propagator_args"]["propagation_early_stop"]
        self.stable_rounds_count = 0

        # Propagation strategies (adapt to the specific use case)
        self.strategies = {
            "initialization": InitialModelPropagation(self.aggregator, self.trainer, self.engine),
            "stable": StableModelPropagation(self.aggregator, self.trainer, self.engine),
        }

    def start(self):
        print_msg_box(
            msg="Starting propagator functionality...\nModel propagation through the network",
            indent=2,
            title="Propagator",
        )

    def get_round(self):
        return self.engine.get_round()

    def update_and_check_neighbors(self, strategy, eligible_neighbors):
        # Update the status of eligible neighbors
        current_status = [n for n in eligible_neighbors]

        # Check if the deque is full and the new status is different from the last one
        if self.status_history and current_status != self.status_history[-1]:
            logging.info(
                f"Status History deque is full and the new status is different from the last one: {list(self.status_history)}"
            )
            self.status_history.append(current_status)
            return True

        # Add the current status to the deque
        logging.info(f"Adding current status to the deque: {current_status}")
        self.status_history.append(current_status)

        # If the deque is full and all elements are the same, stop propagation
        if len(self.status_history) == self.status_history.maxlen and all(
            s == self.status_history[0] for s in self.status_history
        ):
            logging.info(
                f"Propagator exited for {self.status_history.maxlen} equal rounds: {list(self.status_history)}"
            )
            return False

        return True

    def reset_status_history(self):
        self.status_history.clear()

    async def propagate(self, strategy_id: str):
        self.reset_status_history()
        if strategy_id not in self.strategies:
            logging.info(f"Strategy {strategy_id} not found.")
            return False
        if self.get_round() is None:
            logging.info("Propagation halted: round is not set.")
            return False

        strategy = self.strategies[strategy_id]
        logging.info(f"Starting model propagation with strategy: {strategy_id}")

        current_connections = await self.cm.get_addrs_current_connections(only_direct=True)
        eligible_neighbors = [
            neighbor_addr for neighbor_addr in current_connections if strategy.is_node_eligible(neighbor_addr)
        ]
        logging.info(f"Eligible neighbors for model propagation: {eligible_neighbors}")
        if not eligible_neighbors:
            logging.info("Propagation complete: No eligible neighbors.")
            return False

        logging.info("Checking repeated statuses during propagation")
        if not self.update_and_check_neighbors(strategy, eligible_neighbors):
            logging.info("Exiting propagation due to repeated statuses.")
            return False

        model_params, weight = strategy.prepare_model_payload(None)
        if model_params:
            serialized_model = (
                model_params if isinstance(model_params, bytes) else self.trainer.serialize_model(model_params)
            )
        else:
            serialized_model = None

        round_number = -1 if strategy_id == "initialization" else self.get_round()

        for neighbor_addr in eligible_neighbors:
            asyncio.create_task(self.cm.send_model(neighbor_addr, round_number, serialized_model, weight))

        if len(self.aggregator.get_nodes_pending_models_to_aggregate()) >= len(self.aggregator._federation_nodes):
            return False

        await asyncio.sleep(self.interval)
        return True



import asyncio
import inspect
import logging
import threading


class Locker:
    def __init__(self, name, verbose=True, async_lock=False, *args, **kwargs):
        self._name = name
        self._verbose = verbose
        self._async_lock = async_lock

        if async_lock:
            self._lock = asyncio.Lock(*args, **kwargs)
        else:
            self._lock = threading.Lock(*args, **kwargs)

    def acquire(self, *args, **kwargs):
        caller = inspect.stack()[1]
        filename = caller.filename.split("/")[-1]
        lineno = caller.lineno
        if self._verbose:
            if "timeout" in kwargs:
                logging.debug(
                    f"🔒  Acquiring lock [{self._name}] from {filename}:{lineno} with timeout {kwargs['timeout']}"
                )
            else:
                logging.debug(f"🔒  Acquiring lock [{self._name}] from {filename}:{lineno}")
        if self._async_lock:
            raise RuntimeError("Use 'await acquire_async' for acquiring async locks")
        return self._lock.acquire(*args, **kwargs)

    def release(self, *args, **kwargs):
        caller = inspect.stack()[1]
        filename = caller.filename.split("/")[-1]
        lineno = caller.lineno
        if self._verbose:
            logging.debug(f"🔓  Releasing lock [{self._name}] from {filename}:{lineno}")
        self._lock.release()

    def locked(self):
        result = self._lock.locked()
        if self._verbose:
            logging.debug(f"🔐  Lock [{self._name}] is locked? {result}")
        return result

    def __enter__(self):
        if self._async_lock:
            raise RuntimeError("Use 'async with' for acquiring async locks")
        logging.debug(f"🔒  Acquiring lock [{self._name}] using [with] statement")
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._async_lock:
            raise RuntimeError("Use 'async with' for releasing async locks")
        logging.debug(f"🔓  Releasing lock [{self._name}] using [with] statement")
        self.release()

    async def acquire_async(self, *args, **kwargs):
        caller = inspect.stack()[1]
        filename = caller.filename.split("/")[-1]
        lineno = caller.lineno
        if not self._async_lock:
            raise RuntimeError("Use 'acquire' for acquiring non-async locks")
        if self._verbose:
            logging.debug(f"🔒  Acquiring async lock [{self._name}] from {filename}:{lineno}")
        await self._lock.acquire()

    async def release_async(self, *args, **kwargs):
        caller = inspect.stack()[1]
        filename = caller.filename.split("/")[-1]
        lineno = caller.lineno
        if not self._async_lock:
            raise RuntimeError("Use 'release' for releasing non-async locks")
        if self._verbose:
            logging.debug(f"🔓  Releasing async lock [{self._name}] from {filename}:{lineno}")
        self._lock.release()

    async def locked_async(self):
        result = self._lock.locked()
        if self._verbose:
            logging.debug(f"🔐  Async lock [{self._name}] is locked? {result}")

    async def __aenter__(self):
        logging.debug(f"🔒  Acquiring async lock [{self._name}] using [async with] statement")
        await self.acquire_async()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        logging.debug(f"🔓  Releasing async lock [{self._name}] using [async with] statement")
        await self.release_async()



import asyncio
import logging
import time
from typing import TYPE_CHECKING

from nebula.addons.functions import print_msg_box
from nebula.core.utils.locker import Locker

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


class Forwarder:
    def __init__(self, config, cm: "CommunicationsManager"):
        print_msg_box(msg="Starting forwarder module...", indent=2, title="Forwarder module")
        self.config = config
        self.cm = cm
        self.pending_messages = asyncio.Queue()
        self.pending_messages_lock = Locker("pending_messages_lock", verbose=False, async_lock=True)

        self.interval = self.config.participant["forwarder_args"]["forwarder_interval"]
        self.number_forwarded_messages = self.config.participant["forwarder_args"]["number_forwarded_messages"]
        self.messages_interval = self.config.participant["forwarder_args"]["forward_messages_interval"]

    async def start(self):
        asyncio.create_task(self.run_forwarder())

    async def run_forwarder(self):
        if self.config.participant["scenario_args"]["federation"] == "CFL":
            logging.info("🔁  Federation is CFL. Forwarder is disabled...")
            return
        while True:
            # logging.debug(f"🔁  Pending messages: {self.pending_messages.qsize()}")
            start_time = time.time()
            await self.pending_messages_lock.acquire_async()
            await self.process_pending_messages(messages_left=self.number_forwarded_messages)
            await self.pending_messages_lock.release_async()
            sleep_time = max(0, self.interval - (time.time() - start_time))
            await asyncio.sleep(sleep_time)

    async def process_pending_messages(self, messages_left):
        while messages_left > 0 and not self.pending_messages.empty():
            msg, neighbors = await self.pending_messages.get()
            for neighbor in neighbors[:messages_left]:
                if neighbor not in self.cm.connections:
                    continue
                try:
                    logging.debug(f"🔁  Sending message (forwarding) --> to {neighbor}")
                    await self.cm.send_message(neighbor, msg)
                except Exception as e:
                    logging.exception(f"🔁  Error forwarding message to {neighbor}. Error: {e!s}")
                    pass
                await asyncio.sleep(self.messages_interval)
            messages_left -= len(neighbors)
            if len(neighbors) > messages_left:
                logging.debug("🔁  Putting message back in queue for forwarding to the remaining neighbors")
                await self.pending_messages.put((msg, neighbors[messages_left:]))

    async def forward(self, msg, addr_from):
        if self.config.participant["scenario_args"]["federation"] == "CFL":
            logging.info("🔁  Federation is CFL. Forwarder is disabled...")
            return
        try:
            await self.pending_messages_lock.acquire_async()
            current_connections = await self.cm.get_addrs_current_connections(only_direct=True)
            pending_nodes_to_send = [n for n in current_connections if n != addr_from]
            logging.debug(f"🔁  Puting message in queue for forwarding to {pending_nodes_to_send}")
            await self.pending_messages.put((msg, pending_nodes_to_send))
        except Exception as e:
            logging.exception(f"🔁  Error forwarding message. Error: {e!s}")
        finally:
            await self.pending_messages_lock.release_async()



import logging
import os
import platform
import re
import sys

import requests

from nebula import __version__


def check_version():
    # Check version of NEBULA (__version__ is defined in __init__.py) and compare with __version__ in https://raw.githubusercontent.com/CyberDataLab/nebula/main/nebula/__init__.py
    logging.info("Checking NEBULA version...")
    try:
        r = requests.get("https://raw.githubusercontent.com/CyberDataLab/nebula/main/nebula/__init__.py")
        if r.status_code == 200:
            version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', r.text, re.MULTILINE).group(1)
            if version != __version__:
                logging.info(
                    f"Your NEBULA version is {__version__} and the latest version is {version}. Please update your NEBULA version."
                )
                logging.info(
                    "You can update your NEBULA version downloading the latest version from https://github.com/CyberDataLab/nebula"
                )
                sys.exit(0)
            else:
                logging.info(f"Your NEBULA version is {__version__} and it is the latest version.")
    except Exception as e:
        logging.exception(f"Error while checking NEBULA version: {e}")
        sys.exit(0)


def check_environment():
    logging.info(f"NEBULA Platform version: {__version__}")
    # check_version()

    logging.info("======== Running Environment ========")
    logging.info("OS: " + platform.platform())
    logging.info("Hardware: " + platform.machine())
    logging.info("Python version: " + sys.version)

    try:
        import torch

        logging.info("PyTorch version: " + torch.__version__)
    except ImportError:
        logging.info("PyTorch is not installed properly")
    except Exception:
        pass

    logging.info("======== CPU Configuration ========")
    try:
        import psutil

        load1, load5, load15 = psutil.getloadavg()
        cpu_usage = (load15 / os.cpu_count()) * 100

        logging.info(f"The CPU usage is : {cpu_usage:.0f}%")
        logging.info(
            f"Available CPU Memory: {psutil.virtual_memory().available / 1024 / 1024 / 1024:.1f} G / {psutil.virtual_memory().total / 1024 / 1024 / 1024}G"
        )
    except ImportError:
        logging.info("No CPU information available")
    except Exception:
        pass

    if sys.platform == "win32" or sys.platform == "linux":
        logging.info("======== GPU Configuration ========")
        try:
            import pynvml

            pynvml.nvmlInit()
            devices = pynvml.nvmlDeviceGetCount()
            for i in range(devices):
                handle = pynvml.nvmlDeviceGetHandleByIndex(i)
                gpu_percent = pynvml.nvmlDeviceGetUtilizationRates(handle).gpu
                gpu_temp = pynvml.nvmlDeviceGetTemperature(handle, pynvml.NVML_TEMPERATURE_GPU)
                gpu_mem = pynvml.nvmlDeviceGetMemoryInfo(handle)
                gpu_mem_percent = gpu_mem.used / gpu_mem.total * 100
                gpu_power = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000.0
                gpu_clocks = pynvml.nvmlDeviceGetClockInfo(handle, pynvml.NVML_CLOCK_SM)
                gpu_memory_clocks = pynvml.nvmlDeviceGetClockInfo(handle, pynvml.NVML_CLOCK_MEM)
                gpu_utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
                gpu_fan_speed = pynvml.nvmlDeviceGetFanSpeed(handle)
                logging.info(f"GPU{i} percent: {gpu_percent}")
                logging.info(f"GPU{i} temp: {gpu_temp}")
                logging.info(f"GPU{i} mem percent: {gpu_mem_percent}")
                logging.info(f"GPU{i} power: {gpu_power}")
                logging.info(f"GPU{i} clocks: {gpu_clocks}")
                logging.info(f"GPU{i} memory clocks: {gpu_memory_clocks}")
                logging.info(f"GPU{i} utilization: {gpu_utilization.gpu}")
                logging.info(f"GPU{i} fan speed: {gpu_fan_speed}")
        except ImportError:
            logging.info("pynvml module not found, GPU information unavailable")
        except Exception:
            pass
    else:
        logging.info("GPU information unavailable")



__version__ = "0.0.1"
__description__ = "NEBULA: A Platform for Decentralized Federated Learning"
__long_description__ = "NEBULA: A Platform for Decentralized Federated Learning"
__author__ = "Enrique Tomás Martínez Beltrán"
__long_description_content_type__ = "text/markdown"
__keywords__ = "federated learning, decentralized federated learning, machine learning, deep learning, neural networks, collaborative learning"



import logging
from typing import TYPE_CHECKING

from nebula.core.pb import nebula_pb2

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


class MessagesManager:
    def __init__(self, addr, config, cm: "CommunicationsManager"):
        self.addr = addr
        self.config = config
        self.cm = cm

    def generate_discovery_message(self, action, latitude=0.0, longitude=0.0):
        message = nebula_pb2.DiscoveryMessage(
            action=action,
            latitude=latitude,
            longitude=longitude,
        )
        message_wrapper = nebula_pb2.Wrapper()
        message_wrapper.source = self.addr
        message_wrapper.discovery_message.CopyFrom(message)
        data = message_wrapper.SerializeToString()
        return data

    def generate_control_message(self, action, log="Control message"):
        message = nebula_pb2.ControlMessage(
            action=action,
            log=log,
        )
        message_wrapper = nebula_pb2.Wrapper()
        message_wrapper.source = self.addr
        message_wrapper.control_message.CopyFrom(message)
        data = message_wrapper.SerializeToString()
        return data

    def generate_federation_message(self, action, arguments=[], round=None):
        logging.info(f"Building federation message with [Action {action}], arguments {arguments}, and round {round}")

        message = nebula_pb2.FederationMessage(
            action=action,
            arguments=[str(arg) for arg in (arguments or [])],
            round=round,
        )
        message_wrapper = nebula_pb2.Wrapper()
        message_wrapper.source = self.addr
        message_wrapper.federation_message.CopyFrom(message)
        data = message_wrapper.SerializeToString()
        return data

    def generate_model_message(self, round, serialized_model, weight=1):
        message = nebula_pb2.ModelMessage(
            round=round,
            parameters=serialized_model,
            weight=weight,
        )
        message_wrapper = nebula_pb2.Wrapper()
        message_wrapper.source = self.addr
        message_wrapper.model_message.CopyFrom(message)
        data = message_wrapper.SerializeToString()
        return data

    def generate_connection_message(self, action):
        message = nebula_pb2.ConnectionMessage(
            action=action,
        )
        message_wrapper = nebula_pb2.Wrapper()
        message_wrapper.source = self.addr
        message_wrapper.connection_message.CopyFrom(message)
        data = message_wrapper.SerializeToString()
        return data

    def generate_reputation_message(self, reputation):
        message = nebula_pb2.ReputationMessage(
            reputation=reputation,
        )
        message_wrapper = nebula_pb2.Wrapper()
        message_wrapper.source = self.addr
        message_wrapper.reputation_message.CopyFrom(message)
        data = message_wrapper.SerializeToString()
        return data



import numpy as np
import torch

from nebula.core.aggregation.aggregator import Aggregator


class Median(Aggregator):
    """
    Aggregator: Median
    Authors: Dong Yin et al et al.
    Year: 2021
    Note: https://arxiv.org/pdf/1803.01498.pdf
    """

    def __init__(self, config=None, **kwargs):
        super().__init__(config, **kwargs)

    def get_median(self, weights):
        # check if the weight tensor has enough space
        weight_len = len(weights)

        median = 0
        if weight_len % 2 == 1:
            # odd number, return the median
            median, _ = torch.median(weights, 0)
        else:
            # even number, return the mean of median two numbers
            # sort the tensor
            arr_weights = np.asarray(weights)
            nobs = arr_weights.shape[0]
            start = int(nobs / 2) - 1
            end = int(nobs / 2) + 1
            atmp = np.partition(arr_weights, (start, end - 1), 0)
            sl = [slice(None)] * atmp.ndim
            sl[0] = slice(start, end)
            arr_median = np.mean(atmp[tuple(sl)], axis=0)
            median = torch.tensor(arr_median)
        return median

    def run_aggregation(self, models):
        super().run_aggregation(models)

        models = list(models.values())
        models_params = [m for m, _ in models]

        total_models = len(models)

        accum = {layer: torch.zeros_like(param).float() for layer, param in models[-1][0].items()}

        # Calculate the trimmedmean for each parameter
        for layer in accum:
            weight_layer = accum[layer]
            # get the shape of layer tensor
            l_shape = list(weight_layer.shape)

            # get the number of elements of layer tensor
            number_layer_weights = torch.numel(weight_layer)
            # if its 0-d tensor
            if l_shape == []:
                weights = torch.tensor([models_params[j][layer] for j in range(0, total_models)])
                weights = weights.double()
                w = self.get_median(weights)
                accum[layer] = w

            else:
                # flatten the tensor
                weight_layer_flatten = weight_layer.view(number_layer_weights)

                # flatten the tensor of each model
                models_layer_weight_flatten = torch.stack(
                    [models_params[j][layer].view(number_layer_weights) for j in range(0, total_models)],
                    0,
                )

                # get the weight list [w1j,w2j,··· ,wmj], where wij is the jth parameter of the ith local model
                median = self.get_median(models_layer_weight_flatten)
                accum[layer] = median.view(l_shape)
        return accum



import copy
import logging
from collections import OrderedDict

import torch


def cosine_metric2(
    model1: OrderedDict[str, torch.Tensor],
    model2: OrderedDict[str, torch.Tensor],
    similarity: bool = False,
) -> float | None:
    if model1 is None or model2 is None:
        logging.info("Cosine similarity cannot be computed due to missing model")
        return None

    cos_similarities = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].flatten()
            l2 = model2[layer].flatten()
            if l1.shape != l2.shape:
                # Adjust the shape of the smaller layer to match the larger layer
                min_len = min(l1.shape[0], l2.shape[0])
                l1, l2 = l1[:min_len], l2[:min_len]

            cos_sim = torch.nn.functional.cosine_similarity(l1.unsqueeze(0), l2.unsqueeze(0), dim=1)
            cos_similarities.append(cos_sim.item())

    if cos_similarities:
        avg_cos_sim = torch.mean(torch.tensor(cos_similarities))
        # result = torch.clamp(avg_cos_sim, min=0).item()
        # return result
        return avg_cos_sim.item() if similarity else (1 - avg_cos_sim.item())
    else:
        return None


def cosine_metric(model1: OrderedDict, model2: OrderedDict, similarity: bool = False) -> float | None:
    if model1 is None or model2 is None:
        logging.info("Cosine similarity cannot be computed due to missing model")
        return None

    cos_similarities: list = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].to("cpu")
            l2 = model2[layer].to("cpu")
            if l1.shape != l2.shape:
                # Adjust the shape of the smaller layer to match the larger layer
                min_len = min(l1.shape[0], l2.shape[0])
                l1, l2 = l1[:min_len], l2[:min_len]
            cos = torch.nn.CosineSimilarity(dim=l1.dim() - 1)
            cos_mean = torch.mean(cos(l1.float(), l2.float())).mean()
            cos_similarities.append(cos_mean)
        else:
            logging.info(f"Layer {layer} not found in model 2")

    if cos_similarities:
        cos = torch.Tensor(cos_similarities)
        avg_cos = torch.mean(cos)
        relu_cos = torch.nn.functional.relu(avg_cos)  # relu to avoid negative values
        return relu_cos.item() if similarity else (1 - relu_cos.item())
    else:
        return None


def euclidean_metric(
    model1: OrderedDict[str, torch.Tensor],
    model2: OrderedDict[str, torch.Tensor],
    standardized: bool = False,
    similarity: bool = False,
) -> float | None:
    if model1 is None or model2 is None:
        return None

    distances = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].flatten()
            l2 = model2[layer].flatten()
            if standardized:
                l1 = (l1 - l1.mean()) / l1.std()
                l2 = (l2 - l2.mean()) / l2.std()

            distance = torch.norm(l1 - l2, p=2)
            if similarity:
                norm_sum = torch.norm(l1, p=2) + torch.norm(l2, p=2)
                similarity_score = 1 - (distance / norm_sum if norm_sum != 0 else 0)
                distances.append(similarity_score.item())
            else:
                distances.append(distance.item())

    if distances:
        avg_distance = torch.mean(torch.tensor(distances))
        return avg_distance.item()
    else:
        return None


def minkowski_metric(
    model1: OrderedDict[str, torch.Tensor],
    model2: OrderedDict[str, torch.Tensor],
    p: int,
    similarity: bool = False,
) -> float | None:
    if model1 is None or model2 is None:
        return None

    distances = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].flatten()
            l2 = model2[layer].flatten()

            distance = torch.norm(l1 - l2, p=p)
            if similarity:
                norm_sum = torch.norm(l1, p=p) + torch.norm(l2, p=p)
                similarity_score = 1 - (distance / norm_sum if norm_sum != 0 else 0)
                distances.append(similarity_score.item())
            else:
                distances.append(distance.item())

    if distances:
        avg_distance = torch.mean(torch.tensor(distances))
        return avg_distance.item()
    else:
        return None


def manhattan_metric(
    model1: OrderedDict[str, torch.Tensor],
    model2: OrderedDict[str, torch.Tensor],
    similarity: bool = False,
) -> float | None:
    if model1 is None or model2 is None:
        return None

    distances = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].flatten()
            l2 = model2[layer].flatten()

            distance = torch.norm(l1 - l2, p=1)
            if similarity:
                norm_sum = torch.norm(l1, p=1) + torch.norm(l2, p=1)
                similarity_score = 1 - (distance / norm_sum if norm_sum != 0 else 0)
                distances.append(similarity_score.item())
            else:
                distances.append(distance.item())

    if distances:
        avg_distance = torch.mean(torch.tensor(distances))
        return avg_distance.item()
    else:
        return None


def pearson_correlation_metric(
    model1: OrderedDict[str, torch.Tensor],
    model2: OrderedDict[str, torch.Tensor],
    similarity: bool = False,
) -> float | None:
    if model1 is None or model2 is None:
        return None

    correlations = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].flatten()
            l2 = model2[layer].flatten()

            if l1.shape != l2.shape:
                min_len = min(l1.shape[0], l2.shape[0])
                l1, l2 = l1[:min_len], l2[:min_len]

            correlation = torch.corrcoef(torch.stack((l1, l2)))[0, 1]
            if similarity:
                adjusted_similarity = (correlation + 1) / 2
                correlations.append(adjusted_similarity.item())
            else:
                correlations.append(1 - (correlation + 1) / 2)

    if correlations:
        avg_correlation = torch.mean(torch.tensor(correlations))
        return avg_correlation.item()
    else:
        return None


def jaccard_metric(
    model1: OrderedDict[str, torch.Tensor],
    model2: OrderedDict[str, torch.Tensor],
    similarity: bool = False,
) -> float | None:
    if model1 is None or model2 is None:
        return None

    jaccard_scores = []

    for layer in model1:
        if layer in model2:
            l1 = model1[layer].flatten()
            l2 = model2[layer].flatten()

            intersection = torch.sum(torch.min(l1, l2))
            union = torch.sum(torch.max(l1, l2))

            jaccard_sim = intersection / union if union != 0 else 0
            if similarity:
                jaccard_scores.append(jaccard_sim.item())
            else:
                jaccard_scores.append(1 - jaccard_sim.item())

    if jaccard_scores:
        avg_jaccard = torch.mean(torch.tensor(jaccard_scores))
        return avg_jaccard.item()
    else:
        return None


def normalise_layers(untrusted_params, trusted_params):
    trusted_norms = dict([k, torch.norm(trusted_params[k].data.view(-1).float())] for k in trusted_params.keys())

    normalised_params = copy.deepcopy(untrusted_params)

    state_dict = copy.deepcopy(untrusted_params)
    for layer in untrusted_params:
        layer_norm = torch.norm(state_dict[layer].data.view(-1).float())
        scaling_factor = min(layer_norm / trusted_norms[layer], 1)
        logging.debug(f"Layer: {layer} ScalingFactor {scaling_factor}")
        # logging.info("Scaling client {} layer {} with factor {}".format(client, layer, scaling_factor))
        normalised_layer = torch.mul(state_dict[layer], scaling_factor)
        normalised_params[layer] = normalised_layer

    return normalised_params



import numpy
import torch

from nebula.core.aggregation.aggregator import Aggregator


class Krum(Aggregator):
    """
    Aggregator: Krum
    Authors: Peva Blanchard et al.
    Year: 2017
    Note: https://papers.nips.cc/paper/2017/hash/f4b9ec30ad9f68f89b29639786cb62ef-Abstract.html
    """

    def __init__(self, config=None, **kwargs):
        super().__init__(config, **kwargs)

    def run_aggregation(self, models):
        super().run_aggregation(models)

        models = list(models.values())

        accum = {layer: torch.zeros_like(param).float() for layer, param in models[-1][0].items()}
        total_models = len(models)
        distance_list = [0 for i in range(0, total_models)]
        min_index = 0
        min_distance_sum = float("inf")

        for i in range(0, total_models):
            m1, _ = models[i]
            for j in range(0, total_models):
                m2, _ = models[j]
                distance = 0
                if i == j:
                    distance = 0
                else:
                    for layer in m1:
                        l1 = m1[layer]

                        l2 = m2[layer]
                        distance += numpy.linalg.norm(l1 - l2)
                distance_list[i] += distance

            if min_distance_sum > distance_list[i]:
                min_distance_sum = distance_list[i]
                min_index = i
        m, _ = models[min_index]
        for layer in m:
            accum[layer] = accum[layer] + m[layer]

        return accum



import logging
from abc import ABC, abstractmethod

import lightning as pl
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import torch
from torchmetrics import MetricCollection
from torchmetrics.classification import (
    MulticlassAccuracy,
    MulticlassConfusionMatrix,
    MulticlassF1Score,
    MulticlassPrecision,
    MulticlassRecall,
)

from nebula.addons.functions import print_msg_box

matplotlib.use("Agg")
plt.switch_backend("Agg")
from nebula.config.config import TRAINING_LOGGER

logging_training = logging.getLogger(TRAINING_LOGGER)


class NebulaModel(pl.LightningModule, ABC):
    """
    Abstract class for the NEBULA model.

    This class is an abstract class that defines the interface for the NEBULA model.
    """

    def process_metrics(self, phase, y_pred, y, loss=None):
        """
        Calculate and log metrics for the given phase.
        The metrics are calculated in each batch.
        Args:
            phase (str): One of 'Train', 'Validation', or 'Test'
            y_pred (torch.Tensor): Model predictions
            y (torch.Tensor): Ground truth labels
            loss (torch.Tensor, optional): Loss value
        """

        y_pred_classes = torch.argmax(y_pred, dim=1).detach()
        y = y.detach()
        if phase == "Train":
            self.logger.log_data({f"{phase}/Loss": loss.detach()})
            self.train_metrics.update(y_pred_classes, y)
        elif phase == "Validation":
            self.val_metrics.update(y_pred_classes, y)
        elif phase == "Test (Local)":
            self.test_metrics.update(y_pred_classes, y)
            self.cm.update(y_pred_classes, y) if self.cm is not None else None
        elif phase == "Test (Global)":
            self.test_metrics_global.update(y_pred_classes, y)
            self.cm_global.update(y_pred_classes, y) if self.cm_global is not None else None
        else:
            raise NotImplementedError

        del y_pred_classes, y

    def log_metrics_end(self, phase):
        """
        Log metrics for the given phase.
        Args:
            phase (str): One of 'Train', 'Validation', 'Test (Local)', or 'Test (Global)'
            print_cm (bool): Print confusion matrix
            plot_cm (bool): Plot confusion matrix
        """
        if phase == "Train":
            output = self.train_metrics.compute()
        elif phase == "Validation":
            output = self.val_metrics.compute()
        elif phase == "Test (Local)":
            output = self.test_metrics.compute()
        elif phase == "Test (Global)":
            output = self.test_metrics_global.compute()
        else:
            raise NotImplementedError

        output = {
            f"{phase}/{key.replace('Multiclass', '').split('/')[-1]}": value.detach() for key, value in output.items()
        }

        self.logger.log_data(output, step=self.global_number[phase])

        metrics_str = ""
        for key, value in output.items():
            metrics_str += f"{key}: {value:.4f}\n"
        print_msg_box(
            metrics_str,
            indent=2,
            title=f"{phase} Metrics | Epoch: {self.global_number[phase]} | Round: {self.round}",
            logger_name=TRAINING_LOGGER,
        )

    def generate_confusion_matrix(self, phase, print_cm=False, plot_cm=False):
        """
        Generate and plot the confusion matrix for the given phase.
        Args:
            phase (str): One of 'Train', 'Validation', 'Test (Local)', or 'Test (Global)'
        """
        if phase == "Test (Local)":
            if self.cm is None:
                raise ValueError(f"Confusion matrix not available for {phase} phase.")
            cm = self.cm.compute().cpu()
        elif phase == "Test (Global)":
            if self.cm_global is None:
                raise ValueError(f"Confusion matrix not available for {phase} phase.")
            cm = self.cm_global.compute().cpu()
        else:
            raise NotImplementedError

        if print_cm:
            logging_training.info(f"{phase} / Confusion Matrix:\n{cm}")

        if plot_cm:
            cm_numpy = cm.numpy().astype(int)
            classes = [i for i in range(self.num_classes)]
            fig, ax = plt.subplots(figsize=(12, 12))
            sns.heatmap(
                cm_numpy,
                annot=False,
                fmt="",
                cmap="Blues",
                ax=ax,
                xticklabels=classes,
                yticklabels=classes,
                square=True,
            )
            ax.set_xlabel("Predicted labels", fontsize=12)
            ax.set_ylabel("True labels", fontsize=12)
            ax.set_title(f"{phase} Confusion Matrix", fontsize=16)
            plt.xticks(rotation=90, fontsize=6)
            plt.yticks(rotation=0, fontsize=6)
            plt.tight_layout()
            self.logger.log_figure(fig, step=self.round, name=f"{phase}/CM")
            plt.close()

            del cm_numpy, classes, fig, ax

        # Restablecer la matriz de confusión
        if phase == "Test (Local)":
            self.cm.reset()
        else:
            self.cm_global.reset()

        del cm

    def __init__(
        self,
        input_channels=1,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__()

        self.input_channels = input_channels
        self.num_classes = num_classes
        self.learning_rate = learning_rate

        if metrics is None:
            metrics = MetricCollection([
                MulticlassAccuracy(num_classes=num_classes),
                MulticlassPrecision(num_classes=num_classes),
                MulticlassRecall(num_classes=num_classes),
                MulticlassF1Score(num_classes=num_classes),
            ])
        self.train_metrics = metrics.clone(prefix="Train/")
        self.val_metrics = metrics.clone(prefix="Validation/")
        self.test_metrics = metrics.clone(prefix="Test (Local)/")
        self.test_metrics_global = metrics.clone(prefix="Test (Global)/")
        del metrics
        if confusion_matrix is None:
            self.cm = MulticlassConfusionMatrix(num_classes=num_classes)
            self.cm_global = MulticlassConfusionMatrix(num_classes=num_classes)
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)

        # Round counter (number of training-validation-test rounds)
        self.round = 0

        # Epochs counter
        self.global_number = {
            "Train": 0,
            "Validation": 0,
            "Test (Local)": 0,
            "Test (Global)": 0,
        }

        # Communication manager for sending messages from the model (e.g., prototypes, gradients)
        # Model parameters are sent by default using network.propagator
        self.communication_manager = None

    def set_communication_manager(self, communication_manager):
        self.communication_manager = communication_manager

    def get_communication_manager(self):
        if self.communication_manager is None:
            raise ValueError("Communication manager not set.")
        return self.communication_manager

    @abstractmethod
    def forward(self, x):
        """Forward pass of the model."""
        pass

    @abstractmethod
    def configure_optimizers(self):
        """Optimizer configuration."""
        pass

    def step(self, batch, batch_idx, phase):
        """Training/validation/test step."""
        x, y = batch
        y_pred = self.forward(x)
        loss = self.criterion(y_pred, y)
        self.process_metrics(phase, y_pred, y, loss)

        return loss

    def training_step(self, batch, batch_idx):
        """
        Training step for the model.
        Args:
            batch:
            batch_id:

        Returns:
        """
        return self.step(batch, batch_idx=batch_idx, phase="Train")

    def on_train_start(self):
        logging_training.info(f"{'=' * 10} [Training] Started {'=' * 10}")

    def on_train_end(self):
        logging_training.info(f"{'=' * 10} [Training] Done {'=' * 10}")

    def on_train_epoch_end(self):
        self.log_metrics_end("Train")
        self.train_metrics.reset()
        self.global_number["Train"] += 1

    def validation_step(self, batch, batch_idx):
        """
        Validation step for the model.
        Args:
            batch:
            batch_idx:

        Returns:
        """
        return self.step(batch, batch_idx=batch_idx, phase="Validation")

    def on_validation_end(self):
        pass

    def on_validation_epoch_end(self):
        # In general, the validation phase is done in one epoch
        self.log_metrics_end("Validation")
        self.val_metrics.reset()
        self.global_number["Validation"] += 1

    def test_step(self, batch, batch_idx, dataloader_idx=None):
        """
        Test step for the model.
        Args:
            batch:
            batch_idx:

        Returns:
        """
        if dataloader_idx == 0:
            return self.step(batch, batch_idx=batch_idx, phase="Test (Local)")
        else:
            return self.step(batch, batch_idx=batch_idx, phase="Test (Global)")

    def on_test_start(self):
        logging_training.info(f"{'=' * 10} [Testing] Started {'=' * 10}")

    def on_test_end(self):
        logging_training.info(f"{'=' * 10} [Testing] Done {'=' * 10}")

    def on_test_epoch_end(self):
        # In general, the test phase is done in one epoch
        self.log_metrics_end("Test (Local)")
        self.log_metrics_end("Test (Global)")
        self.generate_confusion_matrix("Test (Local)", print_cm=True, plot_cm=True)
        self.generate_confusion_matrix("Test (Global)", print_cm=True, plot_cm=True)
        self.test_metrics.reset()
        self.test_metrics_global.reset()
        self.global_number["Test (Local)"] += 1
        self.global_number["Test (Global)"] += 1

    def on_round_end(self):
        self.round += 1


class NebulaModelStandalone(NebulaModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # Log metrics per epoch
    def on_train_end(self):
        pass

    def on_train_epoch_end(self):
        self.log_metrics_end("Train")
        self.train_metrics.reset()
        # NebulaModel registers training rounds
        # NebulaModelStandalone register the global number of epochs instead of rounds
        self.global_number["Train"] += 1

    def on_validation_end(self):
        pass

    def on_validation_epoch_end(self):
        self.log_metrics_end("Validation")
        self.global_number["Validation"] += 1
        self.val_metrics.reset()

    def on_test_end(self):
        self.global_number["Test (Local)"] += 1
        self.global_number["Test (Global)"] += 1

    def on_test_epoch_end(self):
        self.log_metrics_end("Test (Local)")
        self.log_metrics_end("Test (Global)")
        self.generate_confusion_matrix("Test (Local)", print_cm=True, plot_cm=True)
        self.generate_confusion_matrix("Test (Global)", print_cm=True, plot_cm=True)
        self.test_metrics.reset()
        self.test_metrics_global.reset()



import logging
from copy import deepcopy
from typing import Any

import numpy as np
import torch
from torchmetrics.functional import pairwise_cosine_similarity

# To take into account:
# - Malicious nodes do not train on their own data
# - Malicious nodes aggregate the weights of the other nodes, but not their own
# - The received weights may be the node own weights (aggregated of neighbors), or
#   if the attack is performed specifically for one of the neighbors, it can take
#   its weights only (should be more effective if they are different).


def create_attack(attack_name):
    """
    Function to create an attack object from its name.
    """
    if attack_name == "GLLNeuronInversionAttack":
        return GLLNeuronInversionAttack()
    elif attack_name == "NoiseInjectionAttack":
        return NoiseInjectionAttack()
    elif attack_name == "SwappingWeightsAttack":
        return SwappingWeightsAttack()
    elif attack_name == "DelayerAttack":
        return DelayerAttack()
    else:
        return None


class Attack:
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.attack(*args, **kwds)

    def attack(self, received_weights):
        """
        Function to perform the attack on the received weights. It should return the
        attacked weights.
        """
        raise NotImplementedError


class GLLNeuronInversionAttack(Attack):
    """
    Function to perform neuron inversion attack on the received weights.
    """

    def __init__(self, strength=5.0, perc=1.0):
        super().__init__()
        self.strength = strength
        self.perc = perc

    def attack(self, received_weights):
        logging.info("[GLLNeuronInversionAttack] Performing neuron inversion attack")
        lkeys = list(received_weights.keys())
        logging.info(f"Layer inverted: {lkeys[-2]}")
        received_weights[lkeys[-2]].data = torch.rand(received_weights[lkeys[-2]].shape) * 10000
        return received_weights


class NoiseInjectionAttack(Attack):
    """
    Function to perform noise injection attack on the received weights.
    """

    def __init__(self, strength=10000, perc=1.0):
        super().__init__()
        self.strength = strength
        self.perc = perc

    def attack(self, received_weights):
        logging.info("[NoiseInjectionAttack] Performing noise injection attack")
        lkeys = list(received_weights.keys())
        for k in lkeys:
            logging.info(f"Layer noised: {k}")
            received_weights[k].data += torch.randn(received_weights[k].shape) * self.strength
        return received_weights


class SwappingWeightsAttack(Attack):
    """
    Function to perform swapping weights attack on the received weights. Note that this
    attack performance is not consistent due to its stochasticity.

    Warning: depending on the layer the code may not work (due to reshaping in between),
    or it may be slow (scales quadratically with the layer size).
    Do not apply to last layer, as it would make the attack detectable (high loss
    on malicious node).
    """

    def __init__(self, layer_idx=0):
        super().__init__()
        self.layer_idx = layer_idx

    def attack(self, received_weights):
        logging.info("[SwappingWeightsAttack] Performing swapping weights attack")
        lkeys = list(received_weights.keys())
        wm = received_weights[lkeys[self.layer_idx]]

        # Compute similarity matrix
        sm = torch.zeros((wm.shape[0], wm.shape[0]))
        for j in range(wm.shape[0]):
            sm[j] = pairwise_cosine_similarity(wm[j].reshape(1, -1), wm.reshape(wm.shape[0], -1))

        # Check rows/cols where greedy approach is optimal
        nsort = np.full(sm.shape[0], -1)
        rows = []
        for j in range(sm.shape[0]):
            k = torch.argmin(sm[j])
            if torch.argmin(sm[:, k]) == j:
                nsort[j] = k
                rows.append(j)
        not_rows = np.array([i for i in range(sm.shape[0]) if i not in rows])

        # Ensure the rest of the rows are fully permuted (not optimal, but good enough)
        nrs = deepcopy(not_rows)
        nrs = np.random.permutation(nrs)
        while np.any(nrs == not_rows):
            nrs = np.random.permutation(nrs)
        nsort[not_rows] = nrs
        nsort = torch.tensor(nsort)

        # Apply permutation to weights
        received_weights[lkeys[self.layer_idx]] = received_weights[lkeys[self.layer_idx]][nsort]
        received_weights[lkeys[self.layer_idx + 1]] = received_weights[lkeys[self.layer_idx + 1]][nsort]
        if self.layer_idx + 2 < len(lkeys):
            received_weights[lkeys[self.layer_idx + 2]] = received_weights[lkeys[self.layer_idx + 2]][:, nsort]
        return received_weights


class DelayerAttack(Attack):
    """
    Function to perform delayer attack on the received weights. It delays the
    weights for an indefinite number of rounds.
    """

    def __init__(self):
        super().__init__()
        self.weights = None

    def attack(self, received_weights):
        logging.info("[DelayerAttack] Performing delayer attack")
        if self.weights is None:
            self.weights = deepcopy(received_weights)
        return self.weights



import logging
import random

import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("Agg")
plt.switch_backend("Agg")

import networkx as nx
import numpy as np

from nebula.core.role import Role


class TopologyManager:
    def __init__(
        self,
        scenario_name=None,
        n_nodes=5,
        b_symmetric=True,
        undirected_neighbor_num=5,
        topology=None,
    ):
        self.scenario_name = scenario_name
        if topology is None:
            topology = []
        self.n_nodes = n_nodes
        self.b_symmetric = b_symmetric
        self.undirected_neighbor_num = undirected_neighbor_num
        self.topology = topology
        # Initialize nodes with array of tuples (0,0,0) with size n_nodes
        self.nodes = np.zeros((n_nodes, 3), dtype=np.int32)

        self.b_fully_connected = False
        if self.undirected_neighbor_num < 2:
            raise ValueError("undirected_neighbor_num must be greater than 2")
        # If the number of neighbors is larger than the number of nodes, then the topology is fully connected
        if self.undirected_neighbor_num >= self.n_nodes - 1 and self.b_symmetric:
            self.b_fully_connected = True

    def __getstate__(self):
        # Return the attributes of the class that should be serialized
        return {
            "scenario_name": self.scenario_name,
            "n_nodes": self.n_nodes,
            "topology": self.topology,
            "nodes": self.nodes,
        }

    def __setstate__(self, state):
        # Set the attributes of the class from the serialized state
        self.scenario_name = state["scenario_name"]
        self.n_nodes = state["n_nodes"]
        self.topology = state["topology"]
        self.nodes = state["nodes"]

    def draw_graph(self, plot=False, path=None):
        g = nx.from_numpy_array(self.topology)
        # pos = nx.layout.spectral_layout(g)
        # pos = nx.spring_layout(g, pos=pos, iterations=50)
        pos = nx.spring_layout(g, k=0.15, iterations=20, seed=42)

        fig = plt.figure(num="Network topology", dpi=100, figsize=(6, 6), frameon=False)
        ax = fig.add_axes([0, 0, 1, 1])
        ax.set_xlim([-1.3, 1.3])
        ax.set_ylim([-1.3, 1.3])
        # ax.axis('off')
        labels = {}
        color_map = []
        server = False
        for k in range(self.n_nodes):
            if str(self.nodes[k][2]) == Role.AGGREGATOR:
                color_map.append("orange")
            elif str(self.nodes[k][2]) == Role.SERVER:
                server = True
                color_map.append("green")
            elif str(self.nodes[k][2]) == Role.TRAINER:
                color_map.append("#6182bd")
            elif str(self.nodes[k][2]) == Role.PROXY:
                color_map.append("purple")
            else:
                color_map.append("red")
            labels[k] = f"P{k}\n" + str(self.nodes[k][0]) + ":" + str(self.nodes[k][1])
        # nx.draw_networkx_nodes(g, pos_shadow, node_color='k', alpha=0.5)
        nx.draw_networkx_nodes(g, pos, node_color=color_map, linewidths=2)
        nx.draw_networkx_labels(g, pos, labels, font_size=10, font_weight="bold")
        nx.draw_networkx_edges(g, pos, width=2)
        # plt.margins(0.0)
        roles = [str(i[2]) for i in self.nodes]
        if Role.AGGREGATOR in roles:
            plt.scatter([], [], c="orange", label="Aggregator")
        if Role.SERVER in roles:
            plt.scatter([], [], c="green", label="Server")
        if Role.TRAINER in roles:
            plt.scatter([], [], c="#6182bd", label="Trainer")
        if Role.PROXY in roles:
            plt.scatter([], [], c="purple", label="Proxy")
        if Role.IDLE in roles:
            plt.scatter([], [], c="red", label="Idle")
        # plt.scatter([], [], c="green", label='Central Server')
        # plt.scatter([], [], c="orange", label='Aggregator')
        # plt.scatter([], [], c="#6182bd", label='Trainer')
        # plt.scatter([], [], c="purple", label='Proxy')
        # plt.scatter([], [], c="red", label='Idle')
        plt.legend()
        # import sys
        # if path is None:
        #    if not os.path.exists(f"{sys.path[0]}/logs/{self.scenario_name}"):
        #        os.makedirs(f"{sys.path[0]}/logs/{self.scenario_name}")
        #    plt.savefig(f"{sys.path[0]}/logs/{self.scenario_name}/topology.png", dpi=100, bbox_inches="tight", pad_inches=0)
        # else:
        plt.savefig(f"{path}", dpi=100, bbox_inches="tight", pad_inches=0)
        # plt.gcf().canvas.draw()
        plt.close()

    def generate_topology(self):
        if self.b_fully_connected:
            self.__fully_connected()
            return

        if self.b_symmetric:
            self.__randomly_pick_neighbors_symmetric()
        else:
            self.__randomly_pick_neighbors_asymmetric()

    def generate_server_topology(self):
        self.topology = np.zeros((self.n_nodes, self.n_nodes), dtype=np.float32)
        self.topology[0, :] = 1
        self.topology[:, 0] = 1
        np.fill_diagonal(self.topology, 0)

    def generate_ring_topology(self, increase_convergence=False):
        self.__ring_topology(increase_convergence=increase_convergence)

    def generate_custom_topology(self, topology):
        self.topology = topology

    def get_matrix_adjacency_from_neighbors(self, neighbors):
        matrix_adjacency = np.zeros((self.n_nodes, self.n_nodes), dtype=np.float32)
        for i in range(self.n_nodes):
            for j in range(self.n_nodes):
                if i in neighbors[j]:
                    matrix_adjacency[i, j] = 1
        return matrix_adjacency

    def get_topology(self):
        if self.b_symmetric:
            return self.topology
        else:
            return self.topology

    def get_nodes(self):
        return self.nodes

    @staticmethod
    def get_coordinates(random_geo=True):
        if random_geo:
            if random.randint(0, 1) == 0:
                # España
                bounds = (36.0, 43.0, -9.0, 3.3)  # min_lat, max_lat, min_lon, max_lon
            else:
                # Suiza
                bounds = (45.8, 47.8, 5.9, 10.5)  # min_lat, max_lat, min_lon, max_lon

            min_latitude, max_latitude, min_longitude, max_longitude = bounds
            latitude = random.uniform(min_latitude, max_latitude)
            longitude = random.uniform(min_longitude, max_longitude)

            return latitude, longitude

    def add_nodes(self, nodes):
        self.nodes = nodes

    def update_nodes(self, config_participants):
        self.nodes = config_participants

    def get_node(self, node_idx):
        return self.nodes[node_idx]

    def get_neighbors_string(self, node_idx):
        # logging.info(f"Topology: {self.topology}")
        # logging.info(f"Nodes: {self.nodes}")
        neighbors_data = []
        for i, node in enumerate(self.topology[node_idx]):
            if node == 1:
                neighbors_data.append(self.nodes[i])

        neighbors_data_strings = [f"{i[0]}:{i[1]}" for i in neighbors_data]
        neighbors_data_string = " ".join(neighbors_data_strings)
        logging.info(f"Neighbors of node participant_{node_idx}: {neighbors_data_string}")
        return neighbors_data_string

    def __ring_topology(self, increase_convergence=False):
        topology_ring = np.array(
            nx.to_numpy_matrix(nx.watts_strogatz_graph(self.n_nodes, 2, 0)),
            dtype=np.float32,
        )

        if increase_convergence:
            # Create random links between nodes in topology_ring
            for i in range(self.n_nodes):
                for j in range(self.n_nodes):
                    if topology_ring[i][j] == 0:
                        if random.random() < 0.1:
                            topology_ring[i][j] = 1
                            topology_ring[j][i] = 1

        np.fill_diagonal(topology_ring, 0)
        self.topology = topology_ring

    def __randomly_pick_neighbors_symmetric(self):
        # First generate a ring topology
        topology_ring = np.array(
            nx.to_numpy_matrix(nx.watts_strogatz_graph(self.n_nodes, 2, 0)),
            dtype=np.float32,
        )

        np.fill_diagonal(topology_ring, 0)

        # After, randomly add some links for each node (symmetric)
        # If undirected_neighbor_num is X, then each node has X links to other nodes
        k = int(self.undirected_neighbor_num)
        topology_random_link = np.array(
            nx.to_numpy_matrix(nx.watts_strogatz_graph(self.n_nodes, k, 0)),
            dtype=np.float32,
        )

        # generate symmetric topology
        topology_symmetric = topology_ring.copy()
        for i in range(self.n_nodes):
            for j in range(self.n_nodes):
                if topology_symmetric[i][j] == 0 and topology_random_link[i][j] == 1:
                    topology_symmetric[i][j] = topology_random_link[i][j]

        np.fill_diagonal(topology_symmetric, 0)

        self.topology = topology_symmetric

    def __randomly_pick_neighbors_asymmetric(self):
        # randomly add some links for each node (symmetric)
        k = self.undirected_neighbor_num
        topology_random_link = np.array(
            nx.to_numpy_matrix(nx.watts_strogatz_graph(self.n_nodes, k, 0)),
            dtype=np.float32,
        )

        np.fill_diagonal(topology_random_link, 0)

        # first generate a ring topology
        topology_ring = np.array(
            nx.to_numpy_matrix(nx.watts_strogatz_graph(self.n_nodes, 2, 0)),
            dtype=np.float32,
        )

        np.fill_diagonal(topology_ring, 0)

        for i in range(self.n_nodes):
            for j in range(self.n_nodes):
                if topology_ring[i][j] == 0 and topology_random_link[i][j] == 1:
                    topology_ring[i][j] = topology_random_link[i][j]

        np.fill_diagonal(topology_ring, 0)

        # randomly delete some links
        out_link_set = set()
        for i in range(self.n_nodes):
            len_row_zero = 0
            for j in range(self.n_nodes):
                if topology_ring[i][j] == 0:
                    len_row_zero += 1
            random_selection = np.random.randint(2, size=len_row_zero)
            index_of_zero = 0
            for j in range(self.n_nodes):
                out_link = j * self.n_nodes + i
                if topology_ring[i][j] == 0:
                    if random_selection[index_of_zero] == 1 and out_link not in out_link_set:
                        topology_ring[i][j] = 1
                        out_link_set.add(i * self.n_nodes + j)
                    index_of_zero += 1

        np.fill_diagonal(topology_ring, 0)

        self.topology = topology_ring

    def __fully_connected(self):
        topology_fully_connected = np.array(
            nx.to_numpy_matrix(nx.watts_strogatz_graph(self.n_nodes, self.n_nodes - 1, 0)),
            dtype=np.float32,
        )

        np.fill_diagonal(topology_fully_connected, 0)

        for i in range(self.n_nodes):
            for j in range(self.n_nodes):
                if topology_fully_connected[i][j] != 1:
                    topology_fully_connected[i][j] = 1

        np.fill_diagonal(topology_fully_connected, 0)

        self.topology = topology_fully_connected



import copy
import random

import torch


def labelFlipping(
    dataset,
    indices,
    poisoned_persent=0,
    targeted=False,
    target_label=4,
    target_changed_label=7,
):
    """
    select flipping_persent of labels, and change them to random values.
    Args:
        dataset: the dataset of training data, torch.util.data.dataset like.
        indices: Indices of subsets, list like.
        flipping_persent: The ratio of labels want to change, float like.
    """
    new_dataset = copy.deepcopy(dataset)
    targets = new_dataset.targets.detach().clone()
    num_indices = len(indices)
    # classes = new_dataset.classes
    # class_to_idx = new_dataset.class_to_idx
    # class_list = [class_to_idx[i] for i in classes]
    class_list = set(targets.tolist())
    if targeted == False:
        num_flipped = int(poisoned_persent * num_indices)
        if num_indices == 0:
            return new_dataset
        if num_flipped > num_indices:
            return new_dataset
        flipped_indice = random.sample(indices, num_flipped)

        for i in flipped_indice:
            t = targets[i]
            flipped = torch.tensor(random.sample(class_list, 1)[0])
            while t == flipped:
                flipped = torch.tensor(random.sample(class_list, 1)[0])
            targets[i] = flipped
    else:
        for i in indices:
            if int(targets[i]) == int(target_label):
                targets[i] = torch.tensor(target_changed_label)
    new_dataset.targets = targets
    return new_dataset



import torch
import torchmetrics

from nebula.core.models.nebulamodel import NebulaModel


class SyscallModelSGDOneClassSVM(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.rand(1, input_channels)
        self.learning_rate = learning_rate
        self.nu = 0.1

        self.w = torch.nn.Parameter(torch.zeros(input_channels), requires_grad=True)
        self.rho = torch.nn.Parameter(torch.zeros(1), requires_grad=True)

        self.epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}

    def forward(self, x):
        return torch.matmul(x, self.w) - self.rho

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(self.parameters(), lr=self.learning_rate)
        return optimizer

    def hinge_loss(self, y):
        return torch.mean(torch.clamp(1 - y, min=0))

    def step(self, batch, batch_idx, phase):
        x, labels = batch
        x = x.to(self.device)
        labels = labels.to(self.device)
        y_pred = self.forward(x)
        if phase == "Train":
            loss = 0.5 * torch.sum(self.w**2) + self.nu * self.hinge_loss(y_pred)
            self.log(f"{phase}/Loss", loss, prog_bar=True)
        else:
            y_pred_classes = (y_pred > 0).type(torch.int64)
            loss = torch.nn.functional.binary_cross_entropy_with_logits(y_pred, labels.float())
            self.log(f"{phase}/Loss", loss, prog_bar=True)
            self.log(
                f"{phase}/Accuracy",
                torchmetrics.functional.accuracy(y_pred_classes, labels, task="multiclass"),
                prog_bar=True,
            )
        return loss



import logging


def print_msg_box(msg, indent=1, width=None, title=None, logger_name=None):
    """Print message-box with optional title."""
    if logger_name:
        logger = logging.getLogger(logger_name)
    else:
        logger = logging.getLogger()

    if not isinstance(msg, str):
        raise TypeError("msg parameter must be a string")

    lines = msg.split("\n")
    space = " " * indent
    if not width:
        width = max(map(len, lines))
        if title:
            width = max(width, len(title))
    box = f"\n╔{'═' * (width + indent * 2)}╗\n"  # upper_border
    if title:
        if not isinstance(title, str):
            raise TypeError("title parameter must be a string")
        box += f"║{space}{title:<{width}}{space}║\n"  # title
        box += f"║{space}{'-' * len(title):<{width}}{space}║\n"  # underscore
    box += "".join([f"║{space}{line:<{width}}{space}║\n" for line in lines])
    box += f"╚{'═' * (width + indent * 2)}╝"  # lower_border
    logger.info(box)



import os


class Utils:
    def __init__(self):
        self.init()

    @classmethod
    def check_path(cls, base_path, relative_path):
        full_path = os.path.normpath(os.path.join(base_path, relative_path))
        base_path = os.path.normpath(base_path)

        if not full_path.startswith(base_path):
            raise Exception("Not allowed")
        return full_path



import json
import logging
import os
from logging import FileHandler, Formatter

CYAN = "\x1b[0;36m"
RESET = "\x1b[0m"

TRAINING_LOGGER = "nebula.training"


class Config:
    topology = {}
    participant = {}

    participants = []  # Configuration of each participant (this information is stored only in the controller)
    participants_path = []

    def __init__(self, entity, topology_config_file=None, participant_config_file=None):
        self.entity = entity

        if topology_config_file is not None:
            self.set_topology_config(topology_config_file)

        if participant_config_file is not None:
            self.set_participant_config(participant_config_file)

        if self.participant != {}:
            self.__default_config()
            self.__set_default_logging()
            self.__set_training_logging()

    def __getstate__(self):
        # Return the attributes of the class that should be serialized
        return {"topology": self.topology, "participant": self.participant}

    def __setstate__(self, state):
        # Set the attributes of the class from the serialized state
        self.topology = state["topology"]
        self.participant = state["participant"]

    def get_topology_config(self):
        return json.dumps(self.topology, indent=2)

    def get_participant_config(self):
        return json.dumps(self.participant, indent=2)

    def get_train_logging_config(self):
        # TBD
        pass

    def __default_config(self):
        self.participant["device_args"]["name"] = (
            f"participant_{self.participant['device_args']['idx']}_{self.participant['network_args']['ip']}_{self.participant['network_args']['port']}"
        )
        self.participant["network_args"]["addr"] = (
            f"{self.participant['network_args']['ip']}:{self.participant['network_args']['port']}"
        )

    def __set_default_logging(self):
        experiment_name = self.participant["scenario_args"]["name"]
        self.log_dir = os.path.join(self.participant["tracking_args"]["log_dir"], experiment_name)
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)
        self.log_filename = f"{self.log_dir}/participant_{self.participant['device_args']['idx']}"
        os.makedirs(os.path.dirname(self.log_filename), exist_ok=True)
        (
            console_handler,
            file_handler,
            file_handler_only_debug,
            exp_errors_file_handler,
        ) = self.__setup_logging(self.log_filename)

        level = logging.DEBUG if self.participant["device_args"]["logging"] else logging.CRITICAL
        logging.basicConfig(
            level=level,
            handlers=[
                console_handler,
                file_handler,
                file_handler_only_debug,
                exp_errors_file_handler,
            ],
        )

    def __setup_logging(self, log_filename):
        info_file_format = (
            f"%(asctime)s - {self.participant['device_args']['name']} - [%(filename)s:%(lineno)d] %(message)s"
        )
        debug_file_format = f"%(asctime)s - {self.participant['device_args']['name']} - [%(filename)s:%(lineno)d] %(message)s\n[in %(pathname)s:%(lineno)d]"
        log_console_format = f"{CYAN}%(asctime)s - {self.participant['device_args']['name']} - [%(filename)s:%(lineno)d]{RESET}\n%(message)s"

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.CRITICAL)
        console_handler.setFormatter(Formatter(log_console_format))

        file_handler = FileHandler(f"{log_filename}.log", mode="w", encoding="utf-8")
        file_handler.setLevel(logging.INFO if self.participant["device_args"]["logging"] else logging.CRITICAL)
        file_handler.setFormatter(Formatter(info_file_format))

        file_handler_only_debug = FileHandler(f"{log_filename}_debug.log", mode="w", encoding="utf-8")
        file_handler_only_debug.setLevel(
            logging.DEBUG if self.participant["device_args"]["logging"] else logging.CRITICAL
        )
        file_handler_only_debug.addFilter(lambda record: record.levelno == logging.DEBUG)
        file_handler_only_debug.setFormatter(Formatter(debug_file_format))

        exp_errors_file_handler = FileHandler(f"{log_filename}_error.log", mode="w", encoding="utf-8")
        exp_errors_file_handler.setLevel(
            logging.WARNING if self.participant["device_args"]["logging"] else logging.CRITICAL
        )
        exp_errors_file_handler.setFormatter(Formatter(debug_file_format))

        return (
            console_handler,
            file_handler,
            file_handler_only_debug,
            exp_errors_file_handler,
        )

    def __set_training_logging(self):
        training_log_filename = f"{self.log_filename}_training"
        info_file_format = (
            f"%(asctime)s - {self.participant['device_args']['name']} - [%(filename)s:%(lineno)d] %(message)s"
        )
        log_console_format = f"{CYAN}%(asctime)s - {self.participant['device_args']['name']} - [%(filename)s:%(lineno)d]{RESET}\n%(message)s"
        level = logging.DEBUG if self.participant["device_args"]["logging"] else logging.CRITICAL

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.CRITICAL)
        console_handler.setFormatter(Formatter(log_console_format))

        file_handler = FileHandler(f"{training_log_filename}.log", mode="w", encoding="utf-8")
        file_handler.setLevel(level)
        file_handler.setFormatter(Formatter(info_file_format))

        logger = logging.getLogger(TRAINING_LOGGER)
        logger.setLevel(level)
        logger.addHandler(console_handler)
        logger.addHandler(file_handler)
        logger.propagate = False

        pl_logger = logging.getLogger("lightning.pytorch")
        pl_logger.setLevel(logging.INFO)
        pl_logger.handlers = []
        pl_logger.propagate = False
        pl_logger.addHandler(console_handler)
        pl_logger.addHandler(file_handler)

    def to_json(self):
        # Return participant configuration as a json string
        return json.dumps(self.participant, sort_keys=False, indent=2)

    # Read the configuration file scenario_config.json, and return a dictionary with the configuration
    def set_participant_config(self, participant_config):
        with open(participant_config) as json_file:
            self.participant = json.load(json_file)

    def set_topology_config(self, topology_config_file):
        with open(topology_config_file) as json_file:
            self.topology = json.load(json_file)

    def add_participant_config(self, participant_config):
        with open(participant_config) as json_file:
            self.participants.append(json.load(json_file))

    def set_participants_config(self, participants_config):
        self.participants = []
        self.participants_path = participants_config
        for participant in participants_config:
            self.add_participant_config(participant)

    def add_participants_config(self, participants_config):
        self.participants_path = participants_config
        for participant in participants_config:
            self.add_participant_config(participant)

    def add_neighbor_from_config(self, addr):
        if self.participant != {}:
            if self.participant["network_args"]["neighbors"] == "":
                self.participant["network_args"]["neighbors"] = addr
                self.participant["mobility_args"]["neighbors_distance"][addr] = None
            else:
                if addr not in self.participant["network_args"]["neighbors"]:
                    self.participant["network_args"]["neighbors"] += " " + addr
                    self.participant["mobility_args"]["neighbors_distance"][addr] = None

    def update_neighbors_from_config(self, current_connections, dest_addr):
        final_neighbors = []
        for n in current_connections:
            if n != dest_addr:
                final_neighbors.append(n)

        final_neighbors_string = " ".join(final_neighbors)
        # Update neighbors
        self.participant["network_args"]["neighbors"] = final_neighbors_string
        # Update neighbors location
        self.participant["mobility_args"]["neighbors_distance"] = {
            n: self.participant["mobility_args"]["neighbors_distance"][n]
            for n in final_neighbors
            if n in self.participant["mobility_args"]["neighbors_distance"]
        }
        logging.info(f"Final neighbors: {final_neighbors_string} (config updated))")

    def remove_neighbor_from_config(self, addr):
        if self.participant != {}:
            if self.participant["network_args"]["neighbors"] != "":
                self.participant["network_args"]["neighbors"] = (
                    self.participant["network_args"]["neighbors"].replace(addr, "").replace("  ", " ").strip()
                )

    def reload_config_file(self):
        config_dir = self.participant["tracking_args"]["config_dir"]
        with open(
            f"{config_dir}/participant_{self.participant['device_args']['idx']}.json",
            "w",
        ) as f:
            f.write(self.to_json())



from collections import OrderedDict

import torch
from skimage.util import random_noise


def modelpoison(model: OrderedDict, poisoned_ratio, noise_type="gaussian"):
    """
    Function to add random noise of various types to the model parameter.
    """
    poisoned_model = OrderedDict()
    if type(noise_type) != str:
        noise_type = noise_type[0]

    for layer in model:
        bt = model[layer]
        t = bt.detach().clone()
        single_point = False
        if len(t.shape) == 0:
            t = t.view(-1)
            single_point = True
        # print(t)
        if noise_type == "salt":
            # Replaces random pixels with 1.
            poisoned = torch.tensor(random_noise(t, mode=noise_type, amount=poisoned_ratio))
        elif noise_type == "gaussian":
            # Gaussian-distributed additive noise.
            poisoned = torch.tensor(random_noise(t, mode=noise_type, mean=0, var=poisoned_ratio, clip=True))
        elif noise_type == "s&p":
            # Replaces random pixels with either 1 or low_val, where low_val is 0 for unsigned images or -1 for signed images.
            poisoned = torch.tensor(random_noise(t, mode=noise_type, amount=poisoned_ratio))
        else:
            print("ERROR: poison attack type not supported.")
            poisoned = t
        if single_point:
            poisoned = poisoned[0]
        poisoned_model[layer] = poisoned

    return poisoned_model



import torch

from nebula.core.models.nebulamodel import NebulaModel


class CIFAR10ModelCNN_V2(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.rand(1, 3, 32, 32)
        self.learning_rate = learning_rate
        self.criterion = torch.nn.CrossEntropyLoss()
        self.conv1 = torch.nn.Conv2d(input_channels, 32, 5, padding=2)
        self.bn1 = torch.nn.BatchNorm2d(32)
        self.conv2 = torch.nn.Conv2d(32, 64, 3, padding=1)
        self.bn2 = torch.nn.BatchNorm2d(64)
        self.conv3 = torch.nn.Conv2d(64, 128, 3, padding=1)
        self.bn3 = torch.nn.BatchNorm2d(128)
        self.pool = torch.nn.MaxPool2d(2, 2)
        self.fc1 = torch.nn.Linear(128 * 4 * 4, 512)
        self.fc2 = torch.nn.Linear(512, num_classes)
        self.dropout = torch.nn.Dropout(0.5)

    def forward(self, x):
        x = self.pool(torch.relu(self.bn1(self.conv1(x))))
        x = self.pool(torch.relu(self.bn2(self.conv2(x))))
        x = self.pool(torch.relu(self.bn3(self.conv3(x))))
        x = x.view(-1, 128 * 4 * 4)
        x = torch.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.fc2(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.learning_rate,
            betas=(self.config["beta1"], self.config["beta2"]),
            amsgrad=self.config["amsgrad"],
        )
        return optimizer



import os

from torchvision import transforms
from torchvision.datasets import MNIST


class MNISTDatasetScikit:
    mnist_train = None
    mnist_val = None

    def __init__(self, partition_id=0, partitions_number=1, iid=True):
        self.partition_id = partition_id
        self.partitions_number = partitions_number
        self.iid = iid

        data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
        os.makedirs(data_dir, exist_ok=True)

        if MNISTDatasetScikit.mnist_train is None:
            MNISTDatasetScikit.mnist_train = MNIST(
                data_dir,
                train=True,
                download=True,
                transform=transforms.ToTensor(),
            )
            if not iid:
                sorted_indexes = MNISTDatasetScikit.mnist_train.targets.sort()[1]
                MNISTDatasetScikit.mnist_train.targets = MNISTDatasetScikit.mnist_train.targets[sorted_indexes]
                MNISTDatasetScikit.mnist_train.data = MNISTDatasetScikit.mnist_train.data[sorted_indexes]

        if MNISTDatasetScikit.mnist_val is None:
            MNISTDatasetScikit.mnist_val = MNIST(
                data_dir,
                train=False,
                download=True,
                transform=transforms.ToTensor(),
            )
            if not iid:
                sorted_indexes = MNISTDatasetScikit.mnist_val.targets.sort()[1]
                MNISTDatasetScikit.mnist_val.targets = MNISTDatasetScikit.mnist_val.targets[sorted_indexes]
                MNISTDatasetScikit.mnist_val.data = MNISTDatasetScikit.mnist_val.data[sorted_indexes]

        self.train_set = MNISTDatasetScikit.mnist_train
        self.test_set = MNISTDatasetScikit.mnist_val

    def train_dataloader(self):
        X_train = self.train_set.data.numpy().reshape(-1, 28 * 28)
        y_train = self.train_set.targets.numpy()
        return X_train, y_train

    def test_dataloader(self):
        X_test = self.test_set.data.numpy().reshape(-1, 28 * 28)
        y_test = self.test_set.targets.numpy()
        return X_test, y_test



import lightning as pl
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import torch
import torch.nn.functional as F
from torchmetrics import MetricCollection
from torchmetrics.classification import (
    MulticlassAccuracy,
    MulticlassConfusionMatrix,
    MulticlassF1Score,
    MulticlassPrecision,
    MulticlassRecall,
)

matplotlib.use("Agg")
plt.switch_backend("Agg")
import logging

from nebula.config.config import TRAINING_LOGGER

logging_training = logging.getLogger(TRAINING_LOGGER)


class ContrastiveLoss(torch.nn.Module):
    """
    Contrastive loss function.
    """

    def __init__(self, mu=0.5):
        super().__init__()
        self.mu = mu
        self.cross_entropy_loss = torch.nn.CrossEntropyLoss()

    def forward(self, local_out, global_out, historical_out, labels):
        """
        Calculates the contrastive loss between the local output, global output, and historical output.

        Args:
            local_out (torch.Tensor): The local output tensor of shape (batch_size, embedding_size).
            global_out (torch.Tensor): The global output tensor of shape (batch_size, embedding_size).
            historical_out (torch.Tensor): The historical output tensor of shape (batch_size, embedding_size).
            labels (torch.Tensor): The ground truth labels tensor of shape (batch_size,).

        Returns:
            torch.Tensor: The contrastive loss value.

        Raises:
            ValueError: If the input tensors have different shapes.

        Notes:
            - The contrastive loss is calculated as the difference between the mean cosine similarity of the local output
                with the historical output and the mean cosine similarity of the local output with the global output,
                multiplied by a scaling factor mu.
            - The cosine similarity values represent the similarity between the corresponding vectors in the input tensors.
            Higher values indicate greater similarity, while lower values indicate less similarity.
        """
        if local_out.shape != global_out.shape or local_out.shape != historical_out.shape:
            raise ValueError("Input tensors must have the same shape.")

        # Cross-entropy loss
        ce_loss = self.cross_entropy_loss(local_out, labels)
        # if round > 1:
        # Positive cosine similarity
        pos_cos_sim = F.cosine_similarity(local_out, historical_out, dim=1).mean()

        # Negative cosine similarity
        neg_cos_sim = -F.cosine_similarity(local_out, global_out, dim=1).mean()

        # Combined loss
        contrastive_loss = ce_loss + self.mu * 0.5 * (pos_cos_sim + neg_cos_sim)

        logging_training.debug(
            f"Contrastive loss (mu={self.mu}) with 0.5 of factor: ce_loss: {ce_loss}, pos_cos_sim_local_historical: {pos_cos_sim}, neg_cos_sim_local_global: {neg_cos_sim}, loss: {contrastive_loss}"
        )
        return contrastive_loss
        # else:
        #    logging_training.debug(f"Cross-entropy loss (local model): {ce_loss}")
        #    return ce_loss


class DualAggModel(pl.LightningModule):
    def process_metrics(self, phase, y_pred, y, loss=None, mode="local"):
        """
        Calculate and log metrics for the given phase.
        Args:
            phase (str): One of 'Train', 'Validation', or 'Test'
            y_pred (torch.Tensor): Model predictions
            y (torch.Tensor): Ground truth labels
            loss (torch.Tensor, optional): Loss value
        """

        y_pred_classes = torch.argmax(y_pred, dim=1)
        if phase == "Train":
            if mode == "local":
                output = self.local_train_metrics(y_pred_classes, y)
            elif mode == "historical":
                output = self.historical_train_metrics(y_pred_classes, y)
            elif mode == "global":
                output = self.global_train_metrics(y_pred_classes, y)
        elif phase == "Validation":
            if mode == "local":
                output = self.local_val_metrics(y_pred_classes, y)
            elif mode == "historical":
                output = self.historical_val_metrics(y_pred_classes, y)
            elif mode == "global":
                output = self.global_val_metrics(y_pred_classes, y)
        elif phase == "Test":
            if mode == "local":
                output = self.local_test_metrics(y_pred_classes, y)
            elif mode == "historical":
                output = self.historical_test_metrics(y_pred_classes, y)
            elif mode == "global":
                output = self.global_test_metrics(y_pred_classes, y)
        else:
            raise NotImplementedError
        # print(f"y_pred shape: {y_pred.shape}, y_pred_classes shape: {y_pred_classes.shape}, y shape: {y.shape}")  # Debug print
        output = {
            f"{mode}/{phase}/{key.replace('Multiclass', '').split('/')[-1]}": value for key, value in output.items()
        }
        self.log_dict(output, prog_bar=True, logger=True)

        if self.local_cm is not None and self.historical_cm is not None and self.global_cm is not None:
            if mode == "local":
                self.local_cm.update(y_pred_classes, y)
            elif mode == "historical":
                self.historical_cm.update(y_pred_classes, y)
            elif mode == "global":
                self.global_cm.update(y_pred_classes, y)

    def log_metrics_by_epoch(self, phase, print_cm=False, plot_cm=False, mode="local"):
        """
        Log all metrics at the end of an epoch for the given phase.
        Args:
            phase (str): One of 'Train', 'Validation', or 'Test'
            :param phase:
            :param plot_cm:
        """
        if mode == "local":
            print(f"Epoch end: {mode} {phase}, epoch number: {self.local_epoch_global_number[phase]}")
        elif mode == "historical":
            print(f"Epoch end: {mode} {phase}, epoch number: {self.historical_epoch_global_number[phase]}")
        elif mode == "global":
            print(f"Epoch end: {mode} {phase}, epoch number: {self.global_epoch_global_number[phase]}")

        if phase == "Train":
            if mode == "local":
                output = self.local_train_metrics.compute()
                self.local_train_metrics.reset()
            elif mode == "historical":
                output = self.historical_train_metrics.compute()
                self.historical_train_metrics.reset()
            elif mode == "global":
                output = self.global_train_metrics.compute()
                self.global_train_metrics.reset()
        elif phase == "Validation":
            if mode == "local":
                output = self.local_val_metrics.compute()
                self.local_val_metrics.reset()
            elif mode == "historical":
                output = self.historical_val_metrics.compute()
                self.historical_val_metrics.reset()
            elif mode == "global":
                output = self.global_val_metrics.compute()
                self.global_val_metrics.reset()
        elif phase == "Test":
            if mode == "local":
                output = self.local_test_metrics.compute()
                self.local_test_metrics.reset()
            elif mode == "historical":
                output = self.historical_test_metrics.compute()
                self.historical_test_metrics.reset()
            elif mode == "global":
                output = self.global_test_metrics.compute()
                self.global_test_metrics.reset()
        else:
            raise NotImplementedError

        output = {
            f"{mode}/{phase}Epoch/{key.replace('Multiclass', '').split('/')[-1]}": value
            for key, value in output.items()
        }

        self.log_dict(output, prog_bar=True, logger=True)

        if self.local_cm is not None and self.historical_cm is not None and self.global_cm is not None:
            if mode == "local":
                cm = self.local_cm.compute().cpu()
            elif mode == "historical":
                cm = self.historical_cm.compute().cpu()
            elif mode == "global":
                cm = self.global_cm.compute().cpu()
            print(f"{mode}/{phase}Epoch/CM\n", cm) if print_cm else None
            if plot_cm:
                plt.figure(figsize=(10, 7))
                ax = sns.heatmap(cm.numpy(), annot=True, fmt="d", cmap="Blues")
                ax.set_xlabel("Predicted labels")
                ax.set_ylabel("True labels")
                ax.set_title("Confusion Matrix")
                ax.set_xticks(range(self.num_classes))
                ax.set_yticks(range(self.num_classes))
                ax.xaxis.set_ticklabels([i for i in range(self.num_classes)])
                ax.yaxis.set_ticklabels([i for i in range(self.num_classes)])
                if mode == "local":
                    self.logger.experiment.add_figure(
                        f"{mode}/{phase}Epoch/CM",
                        ax.get_figure(),
                        global_step=self.local_epoch_global_number[phase],
                    )
                elif mode == "historical":
                    self.logger.experiment.add_figure(
                        f"{mode}/{phase}Epoch/CM",
                        ax.get_figure(),
                        global_step=self.historical_epoch_global_number[phase],
                    )
                elif mode == "global":
                    self.logger.experiment.add_figure(
                        f"{mode}/{phase}Epoch/CM",
                        ax.get_figure(),
                        global_step=self.global_epoch_global_number[phase],
                    )
                plt.close()

        if mode == "local":
            self.local_epoch_global_number[phase] += 1
        elif mode == "historical":
            self.historical_epoch_global_number[phase] += 1
        elif mode == "global":
            self.global_epoch_global_number[phase] += 1

    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        mu=0.5,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__()

        self.input_channels = input_channels
        self.num_classes = num_classes
        self.learning_rate = learning_rate
        self.mu = mu

        if metrics is None:
            metrics = MetricCollection([
                MulticlassAccuracy(num_classes=num_classes),
                MulticlassPrecision(num_classes=num_classes),
                MulticlassRecall(num_classes=num_classes),
                MulticlassF1Score(num_classes=num_classes),
            ])

        # Define metrics
        self.local_train_metrics = metrics.clone(prefix="Local/Train/")
        self.local_val_metrics = metrics.clone(prefix="Local/Validation/")
        self.local_test_metrics = metrics.clone(prefix="Local/Test/")

        self.historical_train_metrics = metrics.clone(prefix="Historical/Train/")
        self.historical_val_metrics = metrics.clone(prefix="Historical/Validation/")
        self.historical_test_metrics = metrics.clone(prefix="Historical/Test/")

        self.global_train_metrics = metrics.clone(prefix="Global/Train/")
        self.global_val_metrics = metrics.clone(prefix="Global/Validation/")
        self.global_test_metrics = metrics.clone(prefix="Global/Test/")

        if confusion_matrix is None:
            self.local_cm = MulticlassConfusionMatrix(num_classes=num_classes)
            self.historical_cm = MulticlassConfusionMatrix(num_classes=num_classes)
            self.global_cm = MulticlassConfusionMatrix(num_classes=num_classes)

        # Set seed for reproducibility initialization
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)

        self.local_epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}
        self.historical_epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}
        self.global_epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.rand(1, 3, 32, 32)
        self.learning_rate = learning_rate
        self.criterion = ContrastiveLoss(mu=self.mu)

        # Define layers of the model
        self.model = torch.nn.Sequential(
            torch.nn.Conv2d(input_channels, 16, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(16, 32, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(32, 64, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Flatten(),
            torch.nn.Linear(64 * 4 * 4, 512),
            torch.nn.ReLU(),
            torch.nn.Linear(512, num_classes),
        )

        # Siamese models of the model above
        self.historical_model = torch.nn.Sequential(
            torch.nn.Conv2d(input_channels, 16, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(16, 32, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(32, 64, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Flatten(),
            torch.nn.Linear(64 * 4 * 4, 512),
            torch.nn.ReLU(),
            torch.nn.Linear(512, num_classes),
        )
        self.global_model = torch.nn.Sequential(
            torch.nn.Conv2d(input_channels, 16, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(16, 32, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(32, 64, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Flatten(),
            torch.nn.Linear(64 * 4 * 4, 512),
            torch.nn.ReLU(),
            torch.nn.Linear(512, num_classes),
        )
        # self.historical_model = copy.deepcopy(self.model)
        # self.global_model = copy.deepcopy(self.model)

    def forward(self, x, mode="local"):
        """Forward pass of the model."""
        if mode == "local":
            return self.model(x)
        elif mode == "global":
            return self.global_model(x)
        elif mode == "historical":
            return self.historical_model(x)
        else:
            raise NotImplementedError

    def configure_optimizers(self):
        """ """
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.learning_rate,
            betas=(self.config["beta1"], self.config["beta2"]),
            amsgrad=self.config["amsgrad"],
        )
        return optimizer

    def step(self, batch, batch_idx, phase):
        images, labels = batch
        images = images.to(self.device)
        labels = labels.to(self.device)
        local_out = self.forward(images, mode="local")
        with torch.no_grad():
            historical_out = self.forward(images, mode="historical")
            global_out = self.forward(images, mode="global")

        loss = self.criterion(local_out, global_out, historical_out, labels)

        # Get metrics for each batch and log them
        self.log(f"{phase}/ConstrastiveLoss", loss, prog_bar=True, logger=True)  # Constrastive loss
        self.process_metrics(phase, local_out, labels, loss, mode="local")
        self.process_metrics(phase, historical_out, labels, loss, mode="historical")
        self.process_metrics(phase, global_out, labels, loss, mode="global")

        return loss

    def training_step(self, batch, batch_id):
        """
        Training step for the model.
        Args:
            batch:
            batch_id:

        Returns:
        """
        return self.step(batch, batch_id, "Train")

    def on_train_epoch_end(self):
        self.log_metrics_by_epoch("Train", print_cm=True, plot_cm=True, mode="local")
        self.log_metrics_by_epoch("Train", print_cm=True, plot_cm=True, mode="historical")
        self.log_metrics_by_epoch("Train", print_cm=True, plot_cm=True, mode="global")

    def validation_step(self, batch, batch_idx):
        """
        Validation step for the model.
        Args:
            batch:
            batch_idx:

        Returns:
        """
        return self.step(batch, batch_idx, "Validation")

    def on_validation_epoch_end(self):
        self.log_metrics_by_epoch("Validation", print_cm=True, plot_cm=False, mode="local")
        self.log_metrics_by_epoch("Validation", print_cm=True, plot_cm=False, mode="historical")
        self.log_metrics_by_epoch("Validation", print_cm=True, plot_cm=False, mode="global")

    def test_step(self, batch, batch_idx):
        """
        Test step for the model.
        Args:
            batch:
            batch_idx:

        Returns:
        """
        return self.step(batch, batch_idx, "Test")

    def on_test_epoch_end(self):
        self.log_metrics_by_epoch("Test", print_cm=True, plot_cm=True, mode="local")
        self.log_metrics_by_epoch("Test", print_cm=True, plot_cm=True, mode="historical")
        self.log_metrics_by_epoch("Test", print_cm=True, plot_cm=True, mode="global")

    def save_historical_model(self):
        """
        Save the current local model as the historical model.
        """
        logging_training.info("Copying local model to historical model.")
        self.historical_model.load_state_dict(self.model.state_dict())

    def global_load_state_dict(self, state_dict):
        """
        Load the given state dictionary into the global model.
        Args:
            state_dict (dict): The state dictionary to load into the global model.
        """
        logging_training.info("Loading state dict into global model.")
        adapted_state_dict = self.adapt_state_dict_for_model(state_dict, "model")
        self.global_model.load_state_dict(adapted_state_dict)

    def historical_load_state_dict(self, state_dict):
        """
        Load the given state dictionary into the historical model.
        Args:
            state_dict (dict): The state dictionary to load into the historical model.
        """
        logging_training.info("Loading state dict into historical model.")
        adapted_state_dict = self.adapt_state_dict_for_model(state_dict, "model")
        self.historical_model.load_state_dict(adapted_state_dict)

    def adapt_state_dict_for_model(self, state_dict, model_prefix):
        """
        Adapt the keys in the provided state_dict to match the structure expected by the model.
        """
        new_state_dict = {}
        prefix = f"{model_prefix}."
        for key, value in state_dict.items():
            if key.startswith(prefix):
                # Remove the specific prefix from each key
                new_key = key[len(prefix) :]
                new_state_dict[new_key] = value
        return new_state_dict

    def get_global_model_parameters(self):
        """
        Get the parameters of the global model.
        """
        return self.global_model.state_dict()

    def print_summary(self):
        """
        Print a summary of local, historical and global models to check if they are the same.
        """
        logging_training.info("Local model summary:")
        logging_training.info(self.model)
        logging_training.info("Historical model summary:")
        logging_training.info(self.historical_model)
        logging_training.info("Global model summary:")
        logging_training.info(self.global_model)



import torch

from nebula.core.models.militarysar import _blocks
from nebula.core.models.nebulamodel import NebulaModel


class MilitarySARModelCNN(NebulaModel):
    def __init__(
        self,
        input_channels=2,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.example_input_array = torch.zeros((1, input_channels, 128, 128))
        self.learning_rate = learning_rate
        self.momentum = 0.9
        self.weight_decay = 4e-3
        self.dropout_rate = 0.5
        self.criterion = torch.nn.CrossEntropyLoss()

        self.model = torch.nn.Sequential(
            _blocks.Conv2DBlock(
                shape=[5, 5, self.input_channels, 16],
                stride=1,
                padding="valid",
                activation="relu",
                max_pool=True,
            ),
            _blocks.Conv2DBlock(
                shape=[5, 5, 16, 32],
                stride=1,
                padding="valid",
                activation="relu",
                max_pool=True,
            ),
            _blocks.Conv2DBlock(
                shape=[6, 6, 32, 64],
                stride=1,
                padding="valid",
                activation="relu",
                max_pool=True,
            ),
            _blocks.Conv2DBlock(shape=[5, 5, 64, 128], stride=1, padding="valid", activation="relu"),
            torch.nn.Dropout(p=self.dropout_rate),
            _blocks.Conv2DBlock(shape=[3, 3, 128, self.num_classes], stride=1, padding="valid"),
            torch.nn.Flatten(),
            torch.nn.Linear(360, num_classes),
        )

    def forward(self, x):
        logits = self.model(x)
        return logits

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(
            self.parameters(),
            lr=self.learning_rate,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )
        # optimizer = torch.optim.Adam(
        #     self.parameters(),
        #     lr=self.learning_rate,
        #     weight_decay=self.weight_decay
        # )
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer=optimizer, step_size=50, gamma=0.1)
        # optimizer = torch.optim.Adam(
        #     self.parameters(),
        #     lr=self.learning_rate
        # )
        # lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(
        #     optimizer=optimizer,
        #     milestones=self.lr_step,
        #     gamma=self.lr_decay
        # )
        return [optimizer], [lr_scheduler]



import asyncio
import collections
import hashlib
import logging
import os
import subprocess
import sys
import traceback
from datetime import datetime
from typing import TYPE_CHECKING

import requests

from nebula.addons.mobility import Mobility
from nebula.core.network.connection import Connection
from nebula.core.network.discoverer import Discoverer
from nebula.core.network.forwarder import Forwarder
from nebula.core.network.messages import MessagesManager
from nebula.core.network.propagator import Propagator
from nebula.core.pb import nebula_pb2
from nebula.core.utils.helper import (
    cosine_metric,
    euclidean_metric,
    jaccard_metric,
    manhattan_metric,
    minkowski_metric,
    pearson_correlation_metric,
)
from nebula.core.utils.locker import Locker

if TYPE_CHECKING:
    from nebula.core.engine import Engine


class CommunicationsManager:
    def __init__(self, engine: "Engine"):
        logging.info("🌐  Initializing Communications Manager")
        self._engine = engine
        self.addr = engine.get_addr()
        self.host = self.addr.split(":")[0]
        self.port = int(self.addr.split(":")[1])
        self.config = engine.get_config()
        self.id = str(self.config.participant["device_args"]["idx"])

        self.register_endpoint = f"http://{self.config.participant['scenario_args']['controller']}/nebula/dashboard/{self.config.participant['scenario_args']['name']}/node/register"
        self.wait_endpoint = f"http://{self.config.participant['scenario_args']['controller']}/nebula/dashboard/{self.config.participant['scenario_args']['name']}/node/wait"

        self._connections = {}
        self.connections_lock = Locker(name="connections_lock", async_lock=True)
        self.connections_manager_lock = Locker(name="connections_manager_lock", async_lock=True)
        self.connection_attempt_lock_incoming = Locker(name="connection_attempt_lock_incoming", async_lock=True)
        self.connection_attempt_lock_outgoing = Locker(name="connection_attempt_lock_outgoing", async_lock=True)
        # Pending connections to be established
        self.pending_connections = set()
        self.incoming_connections = {}
        self.outgoing_connections = {}
        self.ready_connections = set()

        self._mm = MessagesManager(addr=self.addr, config=self.config, cm=self)
        self.received_messages_hashes = collections.deque(
            maxlen=self.config.participant["message_args"]["max_local_messages"]
        )
        self.receive_messages_lock = Locker(name="receive_messages_lock", async_lock=True)

        self._discoverer = Discoverer(addr=self.addr, config=self.config, cm=self)
        # self._health = Health(addr=self.addr, config=self.config, cm=self)
        self._forwarder = Forwarder(config=self.config, cm=self)
        self._propagator = Propagator(cm=self)
        self._mobility = Mobility(config=self.config, cm=self)

        # List of connections to reconnect {addr: addr, tries: 0}
        self.connections_reconnect = []
        self.max_connections = 1000
        self.network_engine = None

        self.stop_network_engine = asyncio.Event()
        self.loop = asyncio.get_event_loop()
        max_concurrent_tasks = 5
        self.semaphore_send_model = asyncio.Semaphore(max_concurrent_tasks)

    @property
    def engine(self):
        return self._engine

    @property
    def connections(self):
        return self._connections

    @property
    def mm(self):
        return self._mm

    @property
    def discoverer(self):
        return self._discoverer

    @property
    def health(self):
        return self._health

    @property
    def forwarder(self):
        return self._forwarder

    @property
    def propagator(self):
        return self._propagator

    @property
    def mobility(self):
        return self._mobility

    async def check_federation_ready(self):
        # Check if all my connections are in ready_connections
        logging.info(
            f"🔗  check_federation_ready | Ready connections: {self.ready_connections} | Connections: {self.connections.keys()}"
        )
        if set(self.connections.keys()) == self.ready_connections:
            return True

    async def add_ready_connection(self, addr):
        self.ready_connections.add(addr)

    async def handle_incoming_message(self, data, addr_from):
        try:
            message_wrapper = nebula_pb2.Wrapper()
            message_wrapper.ParseFromString(data)
            source = message_wrapper.source
            logging.debug(f"📥  handle_incoming_message | Received message from {addr_from} with source {source}")
            if source == self.addr:
                return
            if message_wrapper.HasField("discovery_message"):
                if await self.include_received_message_hash(hashlib.md5(data).hexdigest()):
                    await self.forwarder.forward(data, addr_from=addr_from)
                    await self.handle_discovery_message(source, message_wrapper.discovery_message)
            elif message_wrapper.HasField("control_message"):
                await self.handle_control_message(source, message_wrapper.control_message)
            elif message_wrapper.HasField("federation_message"):
                if await self.include_received_message_hash(hashlib.md5(data).hexdigest()):
                    if self.config.participant["device_args"][
                        "proxy"
                    ] or message_wrapper.federation_message.action == nebula_pb2.FederationMessage.Action.Value(
                        "FEDERATION_START"
                    ):
                        await self.forwarder.forward(data, addr_from=addr_from)
                    await self.handle_federation_message(source, message_wrapper.federation_message)
            elif message_wrapper.HasField("model_message"):
                if await self.include_received_message_hash(hashlib.md5(data).hexdigest()):
                    # TODO: Improve the technique. Now only forward model messages if the node is a proxy
                    # Need to update the expected model messages receiving during the round
                    # Round -1 is the initialization round --> all nodes should receive the model
                    if self.config.participant["device_args"]["proxy"] or message_wrapper.model_message.round == -1:
                        await self.forwarder.forward(data, addr_from=addr_from)
                    await self.handle_model_message(source, message_wrapper.model_message)
            elif message_wrapper.HasField("connection_message"):
                await self.handle_connection_message(source, message_wrapper.connection_message)
            else:
                logging.info(f"Unknown handler for message: {message_wrapper}")
        except Exception as e:
            logging.exception(f"📥  handle_incoming_message | Error while processing: {e}")
            logging.exception(traceback.format_exc())

    async def handle_discovery_message(self, source, message):
        logging.info(
            f"🔍  handle_discovery_message | Received [Action {message.action}] from {source} (network propagation)"
        )
        try:
            await self.engine.event_manager.trigger_event(source, message)
        except Exception as e:
            logging.exception(f"🔍  handle_discovery_message | Error while processing: {e}")

    async def handle_control_message(self, source, message):
        logging.info(
            f"🔧  handle_control_message | Received [Action {message.action}] from {source} with log {message.log}"
        )
        try:
            await self.engine.event_manager.trigger_event(source, message)
        except Exception as e:
            logging.exception(
                f"🔧  handle_control_message | Error while processing: {message.action} {message.log} | {e}"
            )

    async def handle_federation_message(self, source, message):
        logging.info(
            f"📝  handle_federation_message | Received [Action {message.action}] from {source} with arguments {message.arguments}"
        )
        try:
            await self.engine.event_manager.trigger_event(source, message)
        except Exception as e:
            logging.exception(
                f"📝  handle_federation_message | Error while processing: {message.action} {message.arguments} | {e}"
            )

    async def handle_model_message(self, source, message):
        logging.info(f"🤖  handle_model_message | Received model from {source} with round {message.round}")
        if self.get_round() is not None:
            await self.engine.get_round_lock().acquire_async()
            current_round = self.get_round()
            await self.engine.get_round_lock().release_async()

            if message.round != current_round and message.round != -1:
                logging.info(
                    f"❗️  handle_model_message | Received a model from a different round | Model round: {message.round} | Current round: {current_round}"
                )
                if message.round > current_round:
                    logging.info(
                        f"🤖  handle_model_message | Saving model from {source} for future round {message.round}"
                    )
                    await self.engine.aggregator.include_next_model_in_buffer(
                        message.parameters,
                        message.weight,
                        source=source,
                        round=message.round,
                    )
                else:
                    logging.info(f"❗️  handle_model_message | Ignoring model from {source} from a previous round")
                return
            if not self.engine.get_federation_ready_lock().locked() and len(self.engine.get_federation_nodes()) == 0:
                logging.info("🤖  handle_model_message | There are no defined federation nodes")
                return
            try:
                # get_federation_ready_lock() is locked when the model is being initialized (first round)
                # non-starting nodes receive the initialized model from the starting node
                if not self.engine.get_federation_ready_lock().locked() or self.engine.get_initialization_status():
                    decoded_model = self.engine.trainer.deserialize_model(message.parameters)
                    if self.config.participant["adaptive_args"]["model_similarity"]:
                        logging.info("🤖  handle_model_message | Checking model similarity")
                        cosine_value = cosine_metric(
                            self.trainer.get_model_parameters(),
                            decoded_model,
                            similarity=True,
                        )
                        euclidean_value = euclidean_metric(
                            self.trainer.get_model_parameters(),
                            decoded_model,
                            similarity=True,
                        )
                        minkowski_value = minkowski_metric(
                            self.trainer.get_model_parameters(),
                            decoded_model,
                            p=2,
                            similarity=True,
                        )
                        manhattan_value = manhattan_metric(
                            self.trainer.get_model_parameters(),
                            decoded_model,
                            similarity=True,
                        )
                        pearson_correlation_value = pearson_correlation_metric(
                            self.trainer.get_model_parameters(),
                            decoded_model,
                            similarity=True,
                        )
                        jaccard_value = jaccard_metric(
                            self.trainer.get_model_parameters(),
                            decoded_model,
                            similarity=True,
                        )
                        with open(
                            f"{self.log_dir}/participant_{self.idx}_similarity.csv",
                            "a+",
                        ) as f:
                            if os.stat(f"{self.log_dir}/participant_{self.idx}_similarity.csv").st_size == 0:
                                f.write(
                                    "timestamp,source_ip,nodes,round,current_round,cosine,euclidean,minkowski,manhattan,pearson_correlation,jaccard\n"
                                )
                            f.write(
                                f"{datetime.now()}, {source}, {message.round}, {current_round}, {cosine_value}, {euclidean_value}, {minkowski_value}, {manhattan_value}, {pearson_correlation_value}, {jaccard_value}\n"
                            )

                    await self.engine.aggregator.include_model_in_buffer(
                        decoded_model,
                        message.weight,
                        source=source,
                        round=message.round,
                    )

                else:
                    if message.round != -1:
                        # Be sure that the model message is from the initialization round (round = -1)
                        logging.info(
                            f"🤖  handle_model_message | Saving model from {source} for future round {message.round}"
                        )
                        await self.engine.aggregator.include_next_model_in_buffer(
                            message.parameters,
                            message.weight,
                            source=source,
                            round=message.round,
                        )
                        return
                    logging.info(f"🤖  handle_model_message | Initializing model (executed by {source})")
                    try:
                        model = self.engine.trainer.deserialize_model(message.parameters)
                        self.engine.trainer.set_model_parameters(model, initialize=True)
                        logging.info("🤖  handle_model_message | Model Parameters Initialized")
                        self.engine.set_initialization_status(True)
                        await (
                            self.engine.get_federation_ready_lock().release_async()
                        )  # Enable learning cycle once the initialization is done
                        try:
                            await (
                                self.engine.get_federation_ready_lock().release_async()
                            )  # Release the lock acquired at the beginning of the engine
                        except RuntimeError:
                            pass
                    except RuntimeError:
                        pass

            except Exception as e:
                logging.exception(f"🤖  handle_model_message | Unknown error adding model: {e}")
                logging.exception(traceback.format_exc())

        else:
            logging.info("🤖  handle_model_message | Tried to add a model while learning is not running")
            if message.round != -1:
                # Be sure that the model message is from the initialization round (round = -1)
                logging.info(f"🤖  handle_model_message | Saving model from {source} for future round {message.round}")
                await self.engine.aggregator.include_next_model_in_buffer(
                    message.parameters,
                    message.weight,
                    source=source,
                    round=message.round,
                )
        return

    async def handle_connection_message(self, source, message):
        try:
            await self.engine.event_manager.trigger_event(source, message)
        except Exception as e:
            logging.exception(f"🔗  handle_connection_message | Error while processing: {message.action} | {e}")

    def get_connections_lock(self):
        return self.connections_lock

    def get_config(self):
        return self.config

    def get_addr(self):
        return self.addr

    def get_round(self):
        return self.engine.get_round()

    async def start(self):
        logging.info("🌐  Starting Communications Manager...")
        await self.deploy_network_engine()

    async def deploy_network_engine(self):
        logging.info("🌐  Deploying Network engine...")
        self.network_engine = await asyncio.start_server(self.handle_connection_wrapper, self.host, self.port)
        self.network_task = asyncio.create_task(self.network_engine.serve_forever(), name="Network Engine")
        logging.info(f"🌐  Network engine deployed at host {self.host} and port {self.port}")

    async def handle_connection_wrapper(self, reader, writer):
        asyncio.create_task(self.handle_connection(reader, writer))

    async def handle_connection(self, reader, writer):
        async def process_connection(reader, writer):
            try:
                addr = writer.get_extra_info("peername")
                connected_node_id = await reader.readline()
                connected_node_id = connected_node_id.decode("utf-8").strip()
                connected_node_port = addr[1]
                if ":" in connected_node_id:
                    connected_node_id, connected_node_port = connected_node_id.split(":")
                connection_addr = f"{addr[0]}:{connected_node_port}"
                direct = await reader.readline()
                direct = direct.decode("utf-8").strip()
                direct = True if direct == "True" else False
                logging.info(
                    f"🔗  [incoming] Connection from {addr} - {connection_addr} [id {connected_node_id} | port {connected_node_port} | direct {direct}] (incoming)"
                )

                if self.id == connected_node_id:
                    logging.info("🔗  [incoming] Connection with yourself is not allowed")
                    writer.write(b"CONNECTION//CLOSE\n")
                    await writer.drain()
                    writer.close()
                    await writer.wait_closed()
                    return

                async with self.connections_manager_lock:
                    if len(self.connections) >= self.max_connections:
                        logging.info("🔗  [incoming] Maximum number of connections reached")
                        logging.info(f"🔗  [incoming] Sending CONNECTION//CLOSE to {addr}")
                        writer.write(b"CONNECTION//CLOSE\n")
                        await writer.drain()
                        writer.close()
                        await writer.wait_closed()
                        return

                    logging.info(f"🔗  [incoming] Connections: {self.connections}")
                    if connection_addr in self.connections:
                        logging.info(f"🔗  [incoming] Already connected with {self.connections[connection_addr]}")
                        logging.info(f"🔗  [incoming] Sending CONNECTION//EXISTS to {addr}")
                        writer.write(b"CONNECTION//EXISTS\n")
                        await writer.drain()
                        writer.close()
                        await writer.wait_closed()
                        return

                    if connection_addr in self.pending_connections:
                        logging.info(f"🔗  [incoming] Connection with {connection_addr} is already pending")
                        if int(self.host.split(".")[3]) < int(addr[0].split(".")[3]):
                            logging.info(
                                f"🔗  [incoming] Closing incoming connection since self.host < host  (from {connection_addr})"
                            )
                            writer.write(b"CONNECTION//CLOSE\n")
                            await writer.drain()
                            writer.close()
                            await writer.wait_closed()
                            return
                        else:
                            logging.info(
                                f"🔗  [incoming] Closing outgoing connection since self.host >= host (from {connection_addr})"
                            )
                            if connection_addr in self.outgoing_connections:
                                out_reader, out_writer = self.outgoing_connections.pop(connection_addr)
                                out_writer.write(b"CONNECTION//CLOSE\n")
                                await out_writer.drain()
                                out_writer.close()
                                await out_writer.wait_closed()

                    logging.info(f"🔗  [incoming] Including {connection_addr} in pending connections")
                    self.pending_connections.add(connection_addr)
                    self.incoming_connections[connection_addr] = (reader, writer)

                logging.info(f"🔗  [incoming] Creating new connection with {addr} (id {connected_node_id})")
                await writer.drain()
                connection = Connection(
                    self,
                    reader,
                    writer,
                    connected_node_id,
                    addr[0],
                    connected_node_port,
                    direct=direct,
                    config=self.config,
                )
                async with self.connections_manager_lock:
                    logging.info(f"🔗  [incoming] Including {connection_addr} in connections")
                    self.connections[connection_addr] = connection
                    logging.info(f"🔗  [incoming] Sending CONNECTION//NEW to {addr}")
                    writer.write(b"CONNECTION//NEW\n")
                    await writer.drain()
                    writer.write(f"{self.id}\n".encode())
                    await writer.drain()
                    await connection.start()

            except Exception as e:
                logging.exception(f"❗️  [incoming] Error while handling connection with {addr}: {e}")
            finally:
                if connection_addr in self.pending_connections:
                    logging.info(
                        f"🔗  [incoming] Removing {connection_addr} from pending connections: {self.pending_connections}"
                    )
                    self.pending_connections.remove(connection_addr)
                if connection_addr in self.incoming_connections:
                    logging.info(
                        f"🔗  [incoming] Removing {connection_addr} from incoming connections: {self.incoming_connections.keys()}"
                    )
                    self.incoming_connections.pop(connection_addr)

        await process_connection(reader, writer)

    async def stop(self):
        logging.info("🌐  Stopping Communications Manager... [Removing connections and stopping network engine]")
        connections = list(self.connections.values())
        for node in connections:
            await node.stop()
        if hasattr(self, "server"):
            self.network_engine.close()
            await self.network_engine.wait_closed()
            self.network_task.cancel()

    async def run_reconnections(self):
        for connection in self.connections_reconnect:
            if connection["addr"] in self.connections:
                connection["tries"] = 0
                logging.info(f"🔗  Node {connection.addr} is still connected!")
            else:
                connection["tries"] += 1
                await self.connect(connection["addr"])

    def verify_connections(self, neighbors):
        # Return True if all neighbors are connected
        if all(neighbor in self.connections for neighbor in neighbors):
            return True
        return False

    async def network_wait(self):
        await self.stop_network_engine.wait()

    async def deploy_additional_services(self):
        logging.info("🌐  Deploying additional services...")
        self._generate_network_conditions()
        await self._forwarder.start()
        # await self._discoverer.start()
        # await self._health.start()
        self._propagator.start()
        await self._mobility.start()

    def _generate_network_conditions(self):
        # TODO: Implement selection of network conditions from frontend
        if self.config.participant["network_args"]["simulation"]:
            interface = self.config.participant["network_args"]["interface"]
            bandwidth = self.config.participant["network_args"]["bandwidth"]
            delay = self.config.participant["network_args"]["delay"]
            delay_distro = self.config.participant["network_args"]["delay-distro"]
            delay_distribution = self.config.participant["network_args"]["delay-distribution"]
            loss = self.config.participant["network_args"]["loss"]
            duplicate = self.config.participant["network_args"]["duplicate"]
            corrupt = self.config.participant["network_args"]["corrupt"]
            reordering = self.config.participant["network_args"]["reordering"]
            logging.info(
                f"🌐  Network simulation is enabled | Interface: {interface} | Bandwidth: {bandwidth} | Delay: {delay} | Delay Distro: {delay_distro} | Delay Distribution: {delay_distribution} | Loss: {loss} | Duplicate: {duplicate} | Corrupt: {corrupt} | Reordering: {reordering}"
            )
            try:
                results = subprocess.run(
                    [
                        "tcset",
                        str(interface),
                        "--rate",
                        str(bandwidth),
                        "--delay",
                        str(delay),
                        "--delay-distro",
                        str(delay_distro),
                        "--delay-distribution",
                        str(delay_distribution),
                        "--loss",
                        str(loss),
                        "--duplicate",
                        str(duplicate),
                        "--corrupt",
                        str(corrupt),
                        "--reordering",
                        str(reordering),
                    ],
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                )
            except Exception as e:
                logging.exception(f"🌐  Network simulation error: {e}")
                return
        else:
            logging.info("🌐  Network simulation is disabled. Using default network conditions...")

    def _reset_network_conditions(self):
        interface = self.config.participant["network_args"]["interface"]
        logging.info("🌐  Resetting network conditions")
        try:
            results = subprocess.run(
                ["tcdel", str(interface), "--all"],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
        except Exception as e:
            logging.exception(f"❗️  Network simulation error: {e}")
            return

    def _set_network_conditions(
        self,
        interface="eth0",
        network="192.168.50.2",
        bandwidth="5Gbps",
        delay="0ms",
        delay_distro="0ms",
        delay_distribution="normal",
        loss="0%",
        duplicate="0%",
        corrupt="0%",
        reordering="0%",
    ):
        logging.info(
            f"🌐  Changing network conditions | Interface: {interface} | Network: {network} | Bandwidth: {bandwidth} | Delay: {delay} | Delay Distro: {delay_distro} | Delay Distribution: {delay_distribution} | Loss: {loss} | Duplicate: {duplicate} | Corrupt: {corrupt} | Reordering: {reordering}"
        )
        try:
            results = subprocess.run(
                [
                    "tcset",
                    str(interface),
                    "--network",
                    str(network) if network is not None else "",
                    "--rate",
                    str(bandwidth),
                    "--delay",
                    str(delay),
                    "--delay-distro",
                    str(delay_distro),
                    "--delay-distribution",
                    str(delay_distribution),
                    "--loss",
                    str(loss),
                    "--duplicate",
                    str(duplicate),
                    "--corrupt",
                    str(corrupt),
                    "--reordering",
                    str(reordering),
                    "--change",
                ],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
        except Exception as e:
            logging.exception(f"❗️  Network simulation error: {e}")
            return

    async def include_received_message_hash(self, hash_message):
        try:
            await self.receive_messages_lock.acquire_async()
            if hash_message in self.received_messages_hashes:
                # logging.info(f"❗️  handle_incoming_message | Ignoring message already received.")
                return False
            self.received_messages_hashes.append(hash_message)
            if len(self.received_messages_hashes) % 10000 == 0:
                logging.info(f"📥  Received {len(self.received_messages_hashes)} messages")
            return True
        except Exception as e:
            logging.exception(f"❗️  handle_incoming_message | Error including message hash: {e}")
            return False
        finally:
            await self.receive_messages_lock.release_async()

    async def send_message_to_neighbors(self, message, neighbors=None, interval=0):
        if neighbors is None:
            current_connections = await self.get_all_addrs_current_connections(only_direct=True)
            neighbors = set(current_connections)
            logging.info(f"Sending message to ALL neighbors: {neighbors}")
        else:
            logging.info(f"Sending message to neighbors: {neighbors}")

        for neighbor in neighbors:
            asyncio.create_task(self.send_message(neighbor, message))
            if interval > 0:
                await asyncio.sleep(interval)

    async def send_message(self, dest_addr, message):
        try:
            conn = self.connections[dest_addr]
            await conn.send(data=message)
        except Exception as e:
            logging.exception(f"❗️  Cannot send message {message} to {dest_addr}. Error: {e!s}")
            await self.disconnect(dest_addr, mutual_disconnection=False)

    async def send_model(self, dest_addr, round, serialized_model, weight=1):
        async with self.semaphore_send_model:
            try:
                conn = self.connections.get(dest_addr)
                if conn is None:
                    logging.info(f"❗️  Connection with {dest_addr} not found")
                    return
                logging.info(
                    f"Sending model to {dest_addr} with round {round}: weight={weight} | size={sys.getsizeof(serialized_model) / (1024** 2) if serialized_model is not None else 0} MB"
                )
                message = self.mm.generate_model_message(round, serialized_model, weight)
                await conn.send(data=message, is_compressed=True)
                logging.info(f"Model sent to {dest_addr} with round {round}")
            except Exception as e:
                logging.exception(f"❗️  Cannot send model to {dest_addr}: {e!s}")
                await self.disconnect(dest_addr, mutual_disconnection=False)

    async def establish_connection(self, addr, direct=True, reconnect=False):
        logging.info(f"🔗  [outgoing] Establishing connection with {addr} (direct: {direct})")

        async def process_establish_connection(addr, direct, reconnect):
            try:
                host = str(addr.split(":")[0])
                port = str(addr.split(":")[1])
                if host == self.host and port == self.port:
                    logging.info("🔗  [outgoing] Connection with yourself is not allowed")
                    return False

                async with self.connections_manager_lock:
                    if addr in self.connections:
                        logging.info(f"🔗  [outgoing] Already connected with {self.connections[addr]}")
                        return False
                    if addr in self.pending_connections:
                        logging.info(f"🔗  [outgoing] Connection with {addr} is already pending")
                        if int(self.host.split(".")[3]) >= int(host.split(".")[3]):
                            logging.info(
                                f"🔗  [outgoing] Closing outgoing connection since self.host >= host (from {addr})"
                            )
                            return False
                        else:
                            logging.info(
                                f"🔗  [outgoing] Closing incoming connection since self.host < host (from {addr})"
                            )
                            if addr in self.incoming_connections:
                                inc_reader, inc_writer = self.incoming_connections.pop(addr)
                                inc_writer.write(b"CONNECTION//CLOSE\n")
                                await inc_writer.drain()
                                inc_writer.close()
                                await inc_writer.wait_closed()

                    self.pending_connections.add(addr)
                    logging.info(f"🔗  [outgoing] Including {addr} in pending connections: {self.pending_connections}")

                logging.info(f"🔗  [outgoing] Openning connection with {host}:{port}")
                reader, writer = await asyncio.open_connection(host, port)
                logging.info(f"🔗  [outgoing] Connection opened with {writer.get_extra_info('peername')}")

                async with self.connections_manager_lock:
                    self.outgoing_connections[addr] = (reader, writer)

                writer.write(f"{self.id}:{self.port}\n".encode())
                await writer.drain()
                writer.write(f"{direct}\n".encode())
                await writer.drain()

                connection_status = await reader.readline()
                connection_status = connection_status.decode("utf-8").strip()

                logging.info(f"🔗  [outgoing] Received connection status {connection_status} (from {addr})")
                logging.info(f"🔗  [outgoing] Connections: {self.connections}")

                if connection_status == "CONNECTION//CLOSE":
                    logging.info(f"🔗  [outgoing] Connection with {addr} closed")
                    if addr in self.pending_connections:
                        logging.info(
                            f"🔗  [outgoing] Removing {addr} from pending connections: {self.pending_connections}"
                        )
                        self.pending_connections.remove(addr)
                    if addr in self.outgoing_connections:
                        logging.info(
                            f"🔗  [outgoing] Removing {addr} from outgoing connections: {self.outgoing_connections.keys()}"
                        )
                        self.outgoing_connections.pop(addr)
                    if addr in self.incoming_connections:
                        logging.info(
                            f"🔗  [outgoing] Removing {addr} from incoming connections: {self.incoming_connections.keys()}"
                        )
                        self.incoming_connections.pop(addr)
                    writer.close()
                    await writer.wait_closed()
                    return False
                elif connection_status == "CONNECTION//PENDING":
                    logging.info(f"🔗  [outgoing] Connection with {addr} is already pending")
                    writer.close()
                    await writer.wait_closed()
                    return False
                elif connection_status == "CONNECTION//EXISTS":
                    logging.info(f"🔗  [outgoing] Already connected {self.connections[addr]}")
                    writer.close()
                    await writer.wait_closed()
                    return True
                elif connection_status == "CONNECTION//NEW":
                    async with self.connections_manager_lock:
                        connected_node_id = await reader.readline()
                        connected_node_id = connected_node_id.decode("utf-8").strip()
                        logging.info(f"🔗  [outgoing] Received connected node id: {connected_node_id} (from {addr})")
                        logging.info(
                            f"🔗  [outgoing] Creating new connection with {host}:{port} (id {connected_node_id})"
                        )
                        connection = Connection(
                            self,
                            reader,
                            writer,
                            connected_node_id,
                            host,
                            port,
                            direct=direct,
                            config=self.config,
                        )
                        self.connections[addr] = connection
                        await connection.start()
                else:
                    logging.info(f"🔗  [outgoing] Unknown connection status {connection_status}")
                    writer.close()
                    await writer.wait_closed()
                    return False

                if reconnect:
                    logging.info(f"🔗  [outgoing] Reconnection check is enabled on node {addr}")
                    self.connections_reconnect.append({"addr": addr, "tries": 0})

                self.config.add_neighbor_from_config(addr)
                return True
            except Exception as e:
                logging.info(f"❗️  [outgoing] Error adding direct connected neighbor {addr}: {e!s}")
                return False
            finally:
                if addr in self.pending_connections:
                    logging.info(f"🔗  [outgoing] Removing {addr} from pending connections: {self.pending_connections}")
                    self.pending_connections.remove(addr)
                if addr in self.outgoing_connections:
                    logging.info(
                        f"🔗  [outgoing] Removing {addr} from outgoing connections: {self.outgoing_connections.keys()}"
                    )
                    self.outgoing_connections.pop(addr)
                if addr in self.incoming_connections:
                    logging.info(
                        f"🔗  [outgoing] Removing {addr} from incoming connections: {self.incoming_connections.keys()}"
                    )
                    self.incoming_connections.pop(addr)

        asyncio.create_task(process_establish_connection(addr, direct, reconnect))

    async def connect(self, addr, direct=True):
        await self.get_connections_lock().acquire_async()
        duplicated = addr in self.connections.keys()
        await self.get_connections_lock().release_async()
        if duplicated:
            if direct:  # Upcoming direct connection
                if not self.connections[addr].get_direct():
                    logging.info(f"🔗  [outgoing] Upgrading non direct connected neighbor {addr} to direct connection")
                    return await self.establish_connection(addr, direct=True, reconnect=False)
                else:  # Upcoming undirected connection
                    logging.info(f"🔗  [outgoing] Already direct connected neighbor {addr}, reconnecting...")
                    return await self.establish_connection(addr, direct=True, reconnect=False)
            else:
                logging.info(f"❗️  Cannot add a duplicate {addr} (undirected connection), already connected")
                return False
        else:
            if direct:
                return await self.establish_connection(addr, direct=True, reconnect=False)
            else:
                return await self.establish_connection(addr, direct=False, reconnect=False)

    async def register(self):
        data = {"node": self.addr}
        logging.info(f"Registering node {self.addr} in the controller")
        response = requests.post(self.register_endpoint, json=data)
        if response.status_code == 200:
            logging.info(f"Node {self.addr} registered successfully in the controller")
        else:
            logging.error(f"Error registering node {self.addr} in the controller")

    async def wait_for_controller(self):
        while True:
            response = requests.get(self.wait_endpoint)
            if response.status_code == 200:
                logging.info("Continue signal received from controller")
                break
            else:
                logging.info("Waiting for controller signal...")
            await asyncio.sleep(1)

    async def disconnect(self, dest_addr, mutual_disconnection=True):
        logging.info(f"Trying to disconnect {dest_addr}")
        if dest_addr not in self.connections:
            logging.info(f"Connection {dest_addr} not found")
            return
        try:
            if mutual_disconnection:
                await self.connections[dest_addr].send(
                    data=self.mm.generate_connection_message(nebula_pb2.ConnectionMessage.Action.DISCONNECT)
                )
                await asyncio.sleep(1)
                self.connections[dest_addr].stop()
        except Exception as e:
            logging.exception(f"❗️  Error while disconnecting {dest_addr}: {e!s}")
        if dest_addr in self.connections:
            logging.info(f"Removing {dest_addr} from connections")
            del self.connections[dest_addr]
        current_connections = await self.get_all_addrs_current_connections(only_direct=True)
        current_connections = set(current_connections)
        logging.info(f"Current connections: {current_connections}")
        self.config.update_neighbors_from_config(current_connections, dest_addr)

    async def get_all_addrs_current_connections(self, only_direct=False, only_undirected=False):
        try:
            await self.get_connections_lock().acquire_async()
            if only_direct:
                return {addr for addr, conn in self.connections.items() if conn.get_direct()}
            elif only_undirected:
                return {addr for addr, conn in self.connections.items() if not conn.get_direct()}
            else:
                return set(self.connections.keys())
        finally:
            await self.get_connections_lock().release_async()

    async def get_addrs_current_connections(self, only_direct=False, only_undirected=False, myself=False):
        current_connections = await self.get_all_addrs_current_connections(
            only_direct=only_direct, only_undirected=only_undirected
        )
        current_connections = set(current_connections)
        if myself:
            current_connections.add(self.addr)
        return current_connections

    async def get_connection_by_addr(self, addr):
        try:
            await self.get_connections_lock().acquire_async()
            for key, conn in self.connections.items():
                if addr in key:
                    return conn
            return None
        except Exception as e:
            logging.exception(f"Error getting connection by address: {e}")
            return None
        finally:
            await self.get_connections_lock().release_async()

    async def get_direct_connections(self):
        try:
            await self.get_connections_lock().acquire_async()
            return {conn for _, conn in self.connections.items() if conn.get_direct()}
        finally:
            await self.get_connections_lock().release_async()

    async def get_undirect_connections(self):
        try:
            await self.get_connections_lock().acquire_async()
            return {conn for _, conn in self.connections.items() if not conn.get_direct()}
        finally:
            await self.get_connections_lock().release_async()

    async def get_nearest_connections(self, top: int = 1):
        try:
            await self.get_connections_lock().acquire_async()
            sorted_connections = sorted(
                self.connections.values(),
                key=lambda conn: (
                    conn.get_neighbor_distance() if conn.get_neighbor_distance() is not None else float("inf")
                ),
            )
            if top == 1:
                return sorted_connections[0]
            else:
                return sorted_connections[:top]
        finally:
            await self.get_connections_lock().release_async()

    def get_ready_connections(self):
        return {addr for addr, conn in self.connections.items() if conn.get_ready()}

    def check_finished_experiment(self):
        return all(
            conn.get_federated_round() == self.config.participant["scenario_args"]["rounds"] - 1
            for conn in self.connections.values()
        )

    def __str__(self):
        return f"Connections: {[str(conn) for conn in self.connections.values()]}"



import json
import logging
import os

from nebula.addons.trustworthiness.pillar import TrustPillar
from nebula.addons.trustworthiness.utils import write_results_json

dirname = os.path.dirname(__file__)

logger = logging.getLogger(__name__)


class TrustMetricManager:
    """
    Manager class to help store the output directory and handle calls from the FL framework.
    """

    def __init__(self):
        self.factsheet_file_nm = "factsheet.json"
        self.eval_metrics_file_nm = "eval_metrics.json"
        self.nebula_trust_results_nm = "nebula_trust_results.json"

    def evaluate(self, scenario, weights, use_weights=False):
        """
        Evaluates the trustworthiness score.

        Args:
            scenario (object): The scenario in whith the trustworthiness will be calculated.
            weights (dict): The desired weghts of the pillars.
            use_weights (bool): True to turn on the weights in the metric config file, default to False.
        """
        # Get scenario name
        scenario_name = scenario[0]
        factsheet_file = os.path.join(dirname, f"files/{scenario_name}/{self.factsheet_file_nm}")
        metrics_cfg_file = os.path.join(dirname, f"configs/{self.eval_metrics_file_nm}")
        results_file = os.path.join(dirname, f"files/{scenario_name}/{self.nebula_trust_results_nm}")

        if not os.path.exists(factsheet_file):
            logger.error(f"{factsheet_file} is missing! Please check documentation.")
            return

        if not os.path.exists(metrics_cfg_file):
            logger.error(f"{metrics_cfg_file} is missing! Please check documentation.")
            return

        with open(factsheet_file) as f, open(metrics_cfg_file) as m:
            factsheet = json.load(f)
            metrics_cfg = json.load(m)
            metrics = metrics_cfg.items()
            input_docs = {"factsheet": factsheet}

            result_json = {"trust_score": 0, "pillars": []}
            final_score = 0
            result_print = []
            for key, value in metrics:
                pillar = TrustPillar(key, value, input_docs, use_weights)
                score, result = pillar.evaluate()
                weight = weights.get(key)
                final_score += weight * score
                result_print.append([key, score])
                result_json["pillars"].append(result)
            final_score = round(final_score, 2)
            result_json["trust_score"] = final_score
            write_results_json(results_file, result_json)



import torch

from nebula.core.models.nebulamodel import NebulaModel


class Sentiment140ModelRNN(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.zeros(1, 1, 28, 28)
        self.learning_rate = learning_rate

        self.embedding_dim = 300
        self.hidden_dim = 256
        self.n_layers = 1
        self.bidirectional = True
        self.output_dim = num_classes

        self.encoder = torch.nn.LSTM(
            self.embedding_dim,
            self.hidden_dim,
            num_layers=self.n_layers,
            bidirectional=self.bidirectional,
            dropout=0.5,
            batch_first=True,
        )
        self.fc = torch.nn.Linear(self.hidden_dim * 2, self.output_dim)
        self.dropout = torch.nn.Dropout(0.5)

        self.criterion = torch.nn.CrossEntropyLoss()

        self.l1 = torch.nn.Linear(28 * 28, 256)
        self.l2 = torch.nn.Linear(256, 128)
        self.l3 = torch.nn.Linear(128, num_classes)

        self.epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}

    def forward(self, x):
        packed_output, (hidden, cell) = self.encoder(x)
        hidden = self.dropout(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1))
        out = self.fc(hidden)
        out = torch.log_softmax(out, dim=1)
        return out

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer
